/*
 * Adaptability exercise
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "support.h"

#include "global.h"
#include "callbacks.h"
#include "translation.h"
#include "keyboard.h"
#include "adaptability.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_tutor_;
extern gchar *consonants;
extern gchar *symbols;
extern gboolean cursor_is_initializing;
extern gboolean cursor_blink;
extern gint tutor_query;
extern gint cursor_pos;
extern gint n_touchs;
extern gint n_errors;
extern gdouble elapsed_time;
extern GTimer *tmr;
extern GdkFont *font;
extern GdkColor black;
extern GdkColor white;

/**********************************************************************
 * Initialize the adapt exercise window.
 */
void
adapt_init ()
{

  gtk_window_set_title (GTK_WINDOW (window_tutor_),
			_("Klavaro - Adaptability"));
  gtk_widget_hide (lookup_widget (window_tutor_, "frame_lesson"));
  gtk_widget_hide (lookup_widget (window_tutor_, "vseparator_tutor"));
  gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_show"));
  gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_other"));

  gtk_label_set_text (GTK_LABEL
		      (lookup_widget (window_tutor_, "label_heading")),
		      _("Adaptability exercises: automating the fingers "
			"responses, typing over all the keyboard."));

  tutor_query = QUERY_INTRO;
  adapt_update ();
}

/**********************************************************************
 * Update what is shown in the tutor window. 
 */
void
adapt_update ()
{
  gchar text[4096];
  GtkText *wg_text;
  GtkEditable *edit_wg;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (lookup_widget (window_tutor_, "text_tutor"));

  gtk_text_freeze (wg_text);
  switch (tutor_query)
    {
    case QUERY_INTRO:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      read_text (text, 4096, "adapt_intro.txt");
      gtk_text_insert (wg_text, font, &black, &white, text, -1);
      tutor_message (_("Press any key to start the exercise. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_START:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      draw_random_pattern ();
      tutor_message (_("Start typing when you are ready. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_PROCESS_TOUCHS:
      tutor_message (_("STARTED! => "));
      gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_END:
      tutor_message (_("<<==  End of exercise."
		       " Press [Enter] to start another. ==>> "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;
    }
  gtk_text_thaw (wg_text);
}

/*
 * Respond to each touch of the user, according to the tutor_query mode
 */
void
adapt_process_touch (gchar user_chr)
{
  GtkText *wg_text;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  switch (tutor_query)
    {
    case QUERY_PROCESS_TOUCHS:
      break;

    case QUERY_INTRO:
      tutor_query = QUERY_START;
      adapt_update ();
      n_touchs = 0;
      n_errors = 0;
      gtk_text_set_point (wg_text, 2);
      cursor_pos = 2;
      if (!cursor_is_initializing)
	{
	  cursor_blink = FALSE;
	  cursor_is_initializing = TRUE;
	  g_timeout_add (TIME_CURSOR_ON + 100, &init_cursor, NULL);
	}
      return;

    case QUERY_START:
      tutor_query = QUERY_PROCESS_TOUCHS;
      adapt_update ();
      g_timer_start (tmr);
      forward_cursor (user_chr);
      return;

    case QUERY_END:
      if (user_chr == '\n')
	{
	  tutor_query = QUERY_INTRO;
	  adapt_process_touch ('\0');
	}
      else
	{
	  tutor_beep ();
	  adapt_update ();
	}
      return;

    default:
      g_warning ("adapt_process_touch(): the program mustn't pass here!");
      return;
    }

  /*
   * It is time to analise the correctness of typing.
   */
  forward_cursor (user_chr);
  if (cursor_pos >= gtk_text_get_length (wg_text))
    {
      elapsed_time = g_timer_elapsed (tmr, NULL);
      adapt_end_exercise ();
      tutor_query = QUERY_END;
      adapt_update ();
      tutor_beep ();
    }
}

/**********************************************************************
 * Put on the screen the final results
 */
#define MAX_TEXT_LEN 1024
void
adapt_end_exercise ()
{
  gchar *tmp_locale;
  gchar *str;
  gchar tmp_str[MAX_TEXT_LEN];
  gchar *tmp_name;
  GtkText *wg_text;
  FILE *fh;
  time_t tmp_time;
  struct tm *ltime;
  gint minutes;
  gint seconds;
  gdouble accuracy;
  gdouble touchs_per_second;
  gdouble words_per_minute;

  /*
   * Calculate statistics
   */
  minutes = ((gulong) elapsed_time) / 60;
  seconds = ((gulong) elapsed_time) % 60;

  accuracy = 100 * (1.0 - (gfloat) n_errors / n_touchs);
  touchs_per_second = (n_touchs - n_errors) / elapsed_time;
  words_per_minute = 10 * touchs_per_second;

  /*
   * Changing to "C" locale
   */
  tmp_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
  setlocale (LC_NUMERIC, "C");

  /*
   * Logging
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/stat_adapt.txt", NULL);
  assert_user_dir ();
  if (!(fh = fopen (tmp_name, "r")))
    {
      fh = fopen (tmp_name, "w");
      fprintf (fh, _("Accuracy\tVelocity\tFluidness\tDate\tHour\tLesson\n"));
    }
  else
    {
      fclose (fh);
      fh = fopen (tmp_name, "a");
    }

  if (fh)
    {
      tmp_time = time (NULL);
      ltime = localtime (&tmp_time);
      fprintf (fh, "%.2f\t%.2f\t%.2f\t%i-%2.2i-%2.2i\t%2.2i:%2.2i\t%s\n",
	       accuracy, words_per_minute, 0.0, (ltime->tm_year) + 1900,
	       (ltime->tm_mon) + 1, (ltime->tm_mday), (ltime->tm_hour),
	       (ltime->tm_min), _("Default"));
      fclose (fh);
    }
  else
    g_message ("not able to log on this file:\n %s", tmp_name);
  g_free (tmp_name);

  /*
   * Coming back to the right locale
   */
  setlocale (LC_NUMERIC, tmp_locale);
  g_free (tmp_locale);

  /*
   * Print statistics
   */
  str = g_strconcat ("\n\n ", _("STATISTICS"),
		     "\n\n ",
		     _("Elapsed time:"), " %i ", _("minute(s) and"), " %i ",
		     _("seconds"), "\n\n ", _("Accuracy:"), " %.1f %%\t\t",
		     _("Goal:"), " %i %%", "\n\n", NULL);
  sprintf (tmp_str, str, minutes, seconds, accuracy, 99);
  g_free (str);

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  gtk_text_insert (wg_text, font, &black, &white, tmp_str, -1);

  /*
   * Comments
   */
  strcpy (tmp_str, _(" Comments:\n\n"));
  str = tmp_str + strlen (tmp_str);
  strcpy (str, "...");
  if (accuracy < 50)
    sprintf (str, _(" You must be kidding.\n"
		    " Your error rate is too high!\n"
		    " Please do it right.\n"));

  else if (accuracy < 90)
    sprintf (str,
	     _(" You are not ready to practice this exercise.\n"
	       " Your accuracy rate is below 90 %%...\n"
	       " Try to accomplish most of the basic course lessons.\n"));
  else if (accuracy < 95)
    sprintf (str,
	     _(" You are doing well. But...\n"
	       " Could you make the accuracy reach at least 95 %%?\n"));
  else if (accuracy < 99)
    sprintf (str,
	     _(" You are almost there,"
	       " but your accuracy rate is still below 99 %%.\n"
	       " Try a few more times,"
	       " so that I can evaluate your velocity.\n"));
  else
    sprintf (str,
	     _(" Very good!\n"
	       " You succeeded with an accuracy rate above 99 %%.\n"
	       " Now it is time to increase your velocity.\n"
	       " Go to the 3rd exercise at the main menu.\n"));

  gtk_text_insert (wg_text, NULL, &black, &white, tmp_str, -1);
}

/**********************************************************************
 * Writes a random pattern of weird words in the exercise window
 */
void
draw_random_pattern ()
{
  gchar *text;
  gchar str_9[10];
  gint i, j;

  for (i = 0; i < 4; i++)
    {				/* 4 paragraphs per exercise */
      text = NULL;
      for (j = 0; j < 22; j++)
	{			/* 22 words per paragraph */
	  if (rand () % 15)
	    create_word (str_9);
	  else
	    create_number (str_9);
	  if (!j)
	    {
	      str_9[0] = toupper (str_9[0]);
	      text = g_strconcat (str_9, NULL);
	    }
	  else
	    text = g_strconcat (text, " ", str_9, NULL);
	}
      text = g_strconcat (text, ".\n", NULL);
      draw_paragraph (text);
      g_free (text);
    }
}

/*
 * Creates a random weird word
 */
void
create_word (gchar str_9[10])
{
  gint i, n, clen, slen;
  const gchar vowels[] = "aeiou";

  clen = strlen (consonants);
  slen = strlen (symbols);
  n = rand () % 8 + 1;
  for (i = 0; i < n; i++)
    {
      if ((rand () % 25))
	{			/* literal */
	  if (i % 2)		/* vowel */
	    if (rand () % 30)
	      str_9[i] = vowels[rand () % 5];
	    else
	      str_9[i] = consonants[rand () % clen];
	  else /* consonant */ if (rand () % 50)
	    str_9[i] = consonants[rand () % clen];
	  else
	    str_9[i] = vowels[rand () % 5];
	  if (!i && !(rand () % 5))	/* capital */
	    str_9[0] = toupper (str_9[0]);
	}
      else
	{			/* symbol */
	  str_9[i] = symbols[rand () % slen];
	  if (rand () % 4)
	    {			/* space after symbol */
	      str_9[i + 1] = '\0';
	      return;
	    }
	}
    }
  if (rand () % 30)
    str_9[n] = vowels[rand () % 5];
  else
    str_9[n] = consonants[rand () % clen];
  str_9[n + 1] = '\0';
}

/*
 * Creates a random number
 */
void
create_number (gchar str_9[10])
{
  if (rand () % 3)
    snprintf (str_9, 10, "%i", rand () % 2500 - 400);
  else
    snprintf (str_9, 10, "%.2f",
	      (rand () % 10000 - 2000) + (rand () % 100) / 100.0);
  str_9[9] = '\0';
}
