/*
 * Basic course
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "support.h"

#include "global.h"
#include "callbacks.h"
#include "translation.h"
#include "keyboard.h"
#include "basic.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_tutor_;
extern gchar *data_path;
extern gchar lowerchars[4][KEY_LINE_LEN + 1];
extern gchar upperchars[4][KEY_LINE_LEN + 1];
extern gint basic_lesson;
extern gint tutor_query;
extern gint cursor_pos;
extern gint n_touchs;
extern gint n_errors;
extern gboolean callbacks_shield;
extern gboolean cursor_is_initializing;
extern gboolean cursor_blink;
extern gdouble elapsed_time;
extern GTimer *tmr;
extern GdkFont *font;
extern GdkColor black;
extern GdkColor white;
extern GdkColor blue;

/**********************************************************************
 * Global variables used only here
 */
gchar lesson_chr_set[6 * 14];

/**********************************************************************
 * Initialize the course and read the user status file to select the
 * last lesson.
 */
void
basic_init ()
{
  gchar *tmp_name;
  FILE *fh;

  gtk_window_set_title (GTK_WINDOW (window_tutor_),
			_("Klavaro - Basic Course"));
  gtk_widget_show (lookup_widget (window_tutor_, "frame_lesson"));
  gtk_widget_show (lookup_widget (window_tutor_, "vseparator_tutor"));
  gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_show"));
  gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_other"));

  /*
   * Retrieve the last lesson where the student had stopped.
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/last_basic_lesson", NULL);
  fh = fopen (tmp_name, "r");
  if (fh)
    {
      fscanf (fh, "%i", &basic_lesson);
      fclose (fh);
      get_lesson_chr_set ();
      tutor_query = QUERY_INTRO;
      basic_process_touch ('\0');
    }
  else
    {
      basic_lesson = 1;
      get_lesson_chr_set ();
      tutor_query = QUERY_INTRO;
      basic_update ();
      gtk_widget_hide (lookup_widget (window_tutor_, "button_revert_lesson"));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_reset_lesson"));

    }
  g_free (tmp_name);
}

/**********************************************************************
 * Update what is shown in the tutor window. 
 */
void
basic_update ()
{
  FILE *fh;
  gint level;
  gchar *tmp_name;
  gchar text[4096];
  GtkLabel *label_wg;
  GtkText *wg_text;
  GtkEditable *edit_wg;

  label_wg = GTK_LABEL (lookup_widget (window_tutor_, "label_heading"));
  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (wg_text);

  gtk_text_freeze (wg_text);
  switch (tutor_query)
    {
    case QUERY_INTRO:
      callbacks_shield = TRUE;
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (lookup_widget
						  (window_tutor_,
						   "spinbutton_lesson")),
				 basic_lesson);
      callbacks_shield = FALSE;

      gtk_label_set_text (label_wg, _("Learning the key positions."));
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      read_text (text, 4096, "basic_intro.txt");
      gtk_text_insert (wg_text, font, &black, &white, text, -1);
      tutor_message (_("Press any key to start the lesson. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_START:
      callbacks_shield = TRUE;
      gtk_spin_button_set_value (GTK_SPIN_BUTTON (lookup_widget
						  (window_tutor_,
						   "spinbutton_lesson")),
				 basic_lesson);
      callbacks_shield = FALSE;

      tmp_name =
	g_strconcat (g_get_home_dir (), "/.klavaro/last_basic_lesson", NULL);
      fh = fopen (tmp_name, "r");
      g_free (tmp_name);
      if (fh)
	{
	  fscanf (fh, "%i", &level);
	  fclose (fh);
	  if (level == basic_lesson)
	    gtk_widget_hide (lookup_widget
			     (window_tutor_, "button_revert_lesson"));
	  else
	    gtk_widget_show (lookup_widget
			     (window_tutor_, "button_revert_lesson"));
	  if (level == 1)
	    gtk_widget_hide (lookup_widget
			     (window_tutor_, "button_reset_lesson"));
	  else
	    gtk_widget_show (lookup_widget
			     (window_tutor_, "button_reset_lesson"));
	}

      snprintf (text, 8 * 14, _("Keys: %s"), lesson_chr_set);
      gtk_label_set_text (label_wg, text);
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      draw_lesson ();
      tutor_message (_("Start typing when you are ready. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_PROCESS_TOUCHS:
      tutor_message (_("STARTED! => "));
      gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_END:
      tutor_message (_("<<==  End of lesson."
		       " Press [Enter] to start another. ==>> "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    default:
      g_warning ("basic_update(): the program mustn't pass here!");
      return;
    }
  gtk_text_thaw (wg_text);
}

/**********************************************************************
 * Respond to each touch of the user, according to the tutor_query mode
 */
void
basic_process_touch (gchar user_chr)
{
  GtkText *wg_text;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  switch (tutor_query)
    {
    case QUERY_PROCESS_TOUCHS:
      break;

    case QUERY_INTRO:
      tutor_query = QUERY_START;
      basic_update ();
      n_touchs = 0;
      n_errors = 0;
      gtk_text_set_point (wg_text, 2);
      cursor_pos = 2;
      if (!cursor_is_initializing)
	{
	  cursor_blink = FALSE;
	  cursor_is_initializing = TRUE;
	  g_timeout_add (TIME_CURSOR_ON + 100, &init_cursor, NULL);
	}
      update_hints (gtk_editable_get_chars (GTK_EDITABLE (wg_text),
					    cursor_pos, cursor_pos + 1)[0]);
      return;

    case QUERY_START:
      tutor_query = QUERY_PROCESS_TOUCHS;
      basic_update ();
      g_timer_start (tmr);
      forward_cursor (user_chr);
      return;

    case QUERY_END:
      if (user_chr == '\n')
	{
	  tutor_query = QUERY_INTRO;
	  basic_process_touch ('\0');
	}
      else
	{
	  tutor_beep ();
	  basic_update ();
	}
      return;

    default:
      g_warning ("basic_process_touch(): the program mustn't pass here!");
      return;
    }

  /*
   * It is time to analise the correctness of typing.
   */
  forward_cursor (user_chr);
  if (cursor_pos >= gtk_text_get_length (wg_text))
    {
      elapsed_time = g_timer_elapsed (tmr, NULL);
      basic_end_lesson ();
      tutor_query = QUERY_END;
      basic_update ();
      tutor_beep ();
      return;
    }
}

/**********************************************************************
 * Put on the screen the final results
 */
#define MAX_TEXT_LEN 1024
void
basic_end_lesson ()
{
  gchar *tmp_locale;
  gchar *str;
  gchar tmp_str[MAX_TEXT_LEN];
  gchar *tmp_name;
  GtkText *wg_text;
  GtkLabel *wg_label;
  FILE *fh;
  time_t tmp_time;
  struct tm *ltime;
  gint minutes;
  gint seconds;
  gdouble accuracy;
  gdouble touchs_per_second;
  gdouble words_per_minute;

  /*
   * Calculate statistics
   */
  minutes = ((gulong) elapsed_time) / 60;
  seconds = ((gulong) elapsed_time) % 60;

  accuracy = 100 * (1.0 - (gfloat) n_errors / n_touchs);
  touchs_per_second = (n_touchs - n_errors) / elapsed_time;
  words_per_minute = 10 * touchs_per_second;

  /*
   * Changing to "C" locale: remember to copy the previous value!
   */
  tmp_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
  setlocale (LC_NUMERIC, "C");

  /*
   * Logging
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/stat_basic.txt", NULL);
  assert_user_dir ();
  if (!(fh = fopen (tmp_name, "r")))
    {
      fh = fopen (tmp_name, "w");
      fprintf (fh, _("Accuracy\tVelocity\tFluidness\tDate\tHour\tLesson\n"));
    }
  else
    {
      fclose (fh);
      fh = fopen (tmp_name, "a");
    }

  if (fh)
    {
      tmp_time = time (NULL);
      ltime = localtime (&tmp_time);
      fprintf (fh, "%.2f\t%.2f\t%.2f\t%i-%2.2i-%2.2i\t%2.2i:%2.2i\t%2.2i\n",
	       accuracy, words_per_minute, 0.0, (ltime->tm_year) + 1900,
	       (ltime->tm_mon) + 1, (ltime->tm_mday), (ltime->tm_hour),
	       (ltime->tm_min), basic_lesson);
      fclose (fh);
    }
  else
    g_message ("not able to log on this file:\n %s", tmp_name);
  g_free (tmp_name);

  /*
   * Coming back to the right locale
   */
  setlocale (LC_NUMERIC, tmp_locale);
  g_free (tmp_locale);

  /*
   * Print statistics
   */
  str = g_strconcat ("\n\n ", _("STATISTICS"),
		     "\n\n ",
		     _("Elapsed time:"), " %i ", _("minute(s) and"), " %i ",
		     _("seconds"), "\n\n ", _("Accuracy:"), " %.1f %%\t\t",
		     _("Goal:"), " %i %%", "\n\n", NULL);
  sprintf (tmp_str, str, minutes, seconds, accuracy, 95);
  g_free (str);

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  gtk_text_insert (wg_text, font, &black, &white, tmp_str, -1);

  /*
   * Comments
   */
  strcpy (tmp_str, _(" Comments:\n\n"));
  str = tmp_str + strlen (tmp_str);
  strcpy (str, "...");
  if (accuracy < 95.0)
    sprintf (str,
	     _(" You did not succeed.\n"
	       " The accuracy must be above 95 %%.\n" " Try again!\n"));
  else
    {
      if (++basic_lesson > 43)
	{
	  basic_lesson = 1;
	  get_lesson_chr_set ();
	  wg_label =
	    GTK_LABEL (lookup_widget (window_tutor_, "label_heading"));
	  gtk_label_set_text (wg_label,
			      _("Positions of keys seems to be learned!"));
	  sprintf (str,
		   _(" Congratulations!\n"
		     " You have accomplished the entire basic course.\n"
		     " Go to the next type of exercise: adaptability.\n"
		     " There you will practice mainly the accuracy.\n"));
	}
      else
	sprintf (str,
		 _(" All right, now you got it!\n"
		   " Go to the next lesson.\n"));

      assert_user_dir ();
      tmp_name =
	g_strconcat (g_get_home_dir (), "/.klavaro/last_basic_lesson", NULL);
      fh = fopen (tmp_name, "w");
      if (!fh)
	g_message
	  ("could not write to the file: ~/.klavaro/last_basic_lesson");
      else
	{
	  fprintf (fh, "%i", basic_lesson);
	  fclose (fh);
	}
      g_free (tmp_name);
    }
  gtk_text_insert (wg_text, NULL, &black, &white, tmp_str, -1);
}

/**********************************************************************
 * Read the characters to be used with the current basic_lesson
 */
gint
get_lesson_chr_set ()
{
  gint i, j, k;
  gchar line_str[16];
  gchar *lesson_file;
  FILE *fh;

  /*
   * Open the lesson file
   */
  lesson_file = g_strconcat (data_path, "basic_lessons.txt", NULL);
  fh = fopen (lesson_file, "r");
  g_free (lesson_file);
  if (!fh)
    g_error ("couldn't find the basic lessons' file.");

  /*
   * Search the lesson
   */
  for (i = 1; i < basic_lesson; i++)
    for (j = 0; j < 11; j++)
      if (!(fgets (line_str, 16, fh)))
	break;

  /*
   * Pass heading line
   */
  if (!(fgets (line_str, 16, fh)))
    {
      lesson_chr_set[0] = '\0';
      fclose (fh);
      return (1);
    }

  /*
   * Get chars, lower set
   */
  for (k = 0, i = 0; i < 4; i++)
    {
      if (!(fgets (line_str, 16, fh)))
	{
	  lesson_chr_set[0] = '\0';
	  fclose (fh);
	  return (-1);
	}
      for (j = 0; j < 14; j++)
	if (line_str[j] == '1' && isgraph (lowerchars[i][j]))
	  lesson_chr_set[k++] = tolower (lowerchars[i][j]);
    }

  /*
   * Pass blank line
   */
  if (!(fgets (line_str, 16, fh)))
    {
      lesson_chr_set[0] = '\0';
      fclose (fh);
      return (-1);
    }

  /*
   * Get chars, upper set
   */
  for (i = 0; i < 4; i++)
    {
      if (!(fgets (line_str, 16, fh)))
	{
	  lesson_chr_set[0] = '\0';
	  fclose (fh);
	  return (-1);
	}
      for (j = 0; j < 14; j++)
	if (line_str[j] == '1' && isgraph (upperchars[i][j]))
	  lesson_chr_set[k++] = tolower (upperchars[i][j]);
    }
  fclose (fh);

  lesson_chr_set[k] = '\0';
  return (0);
}

/**********************************************************************
 * Put the lesson's characters at the main screen
 */
#define N_LINES 8
void
draw_lesson ()
{
  gint i, j, k, len;
  gint remain, idx, rnd;
  gchar sentence[9 * 6 + 1];
  gchar char_pool[N_LINES * 9 * 5];
  GtkText *wg_text;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  gtk_text_set_point (wg_text, 0);
  gtk_text_insert (wg_text, font, &black, &white, "\n ", -1);

  /*
   * Fill in the character pool
   */
  if ((len = strlen (lesson_chr_set)) < 2)
    {
      gtk_text_insert (wg_text, font, &black, &blue, " !!!", -1);
      return;
    }
  for (k = 0; k < (N_LINES * 9 * 5 - len); k += len)
    {
      memmove (char_pool + k, lesson_chr_set, len);
    }
  remain = (N_LINES * 9 * 5) % len;
  remain = remain ? remain : len;
  memmove (char_pool + k, lesson_chr_set, remain);

  /*
   * Draw the lines as sentences
   */
  sentence[9 * 6] = '\0';
  len = N_LINES * 9 * 5;
  for (i = 0; i < N_LINES; i++)
    {				/* lines (sentences) */
      idx = 0;
      for (j = 0; j < 9; j++)
	{			/* words */
	  for (k = 0; k < 5; k++)
	    {			/* letters */
	      rnd = rand () % len;
	      sentence[idx++] = char_pool[rnd];
	      char_pool[rnd] = char_pool[--len];
	    }
	  sentence[idx++] = (j < 8) ? ' ' : 182;
	}
      gtk_text_insert (wg_text, font, &black, &blue, sentence, -1);
      gtk_text_insert (wg_text, font, &black, &white, "\n\n ", -1);
    }
}
