/*
 * The base of the program is here.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <gtkextra/gtkextra.h>

#include "support.h"
#include "interface.h"

#include "global.h"
#include "callbacks.h"
#include "keyboard.h"
#include "translation.h"
#include "basic.h"
#include "adaptability.h"
#include "velocity.h"
#include "fluidness.h"

/**********************************************************************
 * Wide global variables
 */
extern GtkWidget *window_main_;
extern GtkWidget *window_hints_;
extern gboolean callbacks_shield;
extern GdkFont *font;
extern GdkColor black;
extern GdkColor white;
extern GdkColor darkgreen;
extern GdkColor blue;

/**********************************************************************
 * Keyboard variables
 */
extern GtkWidget *window_keyboard_;
extern gboolean is_initializing_list;
extern gboolean keyboard_is_modified;
extern gchar *kb_name;

/**********************************************************************
 * Shared tutor window variables
 */
extern GtkWidget *window_tutor_;
extern GtkWidget *popup_other_;
extern GtkWidget *window_stat_;
extern GTimer *tmr_fast;
extern gint exercise_type;
extern gint tutor_query;
extern gint basic_lesson;
extern gboolean tutor_pause;
extern gboolean cursor_blink;
extern gint cursor_pos;
extern gchar *entry_mesg_buffer;

/**********************************************************************
 * Part 1
 * Callbacks from the main window
 */
/**********************************************************************
 * Main window: initialization and destruction
 */
void
on_window_main_realize (GtkWidget * widget, gpointer user_data)
{
  if (callbacks_shield)
    return;
  initialize_global_variables ();
}

void
on_window_main_destroy (GtkObject * object, gpointer user_data)
{
  if (callbacks_shield)
    return;
  g_message ("adiaux!");
  gtk_exit (0);
}

/**********************************************************************
 * Info-dialog shared tasks
 */
void
on_button_info_return_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

void
on_dialog_info_destroy (GtkObject * object, gpointer user_data)
{
  gtk_widget_show (window_main_);
}


/**********************************************************************
 * Help dialog
 */
void
on_button_help_clicked (GtkButton * button, gpointer user_data)
{
  gchar text[TEXT_MAXLEN];
  GtkWidget *dialog;
  GtkLabel *wg_label;
  GtkText *wg_text;

  dialog = create_dialog_info ();

  wg_label = GTK_LABEL (lookup_widget (dialog, "label_info"));
  gtk_label_set_text (wg_label, _("Help?"));

  wg_text = GTK_TEXT (lookup_widget (dialog, "text_info"));
  read_text (text, TEXT_MAXLEN, "help.txt");
  gtk_text_insert (wg_text, font, &black, &white, text, -1);

  gtk_widget_hide (window_main_);
  gtk_widget_show (dialog);
}

/**********************************************************************
 * About dialog
 */
void
on_button_about_clicked (GtkButton * button, gpointer user_data)
{
  gchar text[TEXT_MAXLEN];
  gchar *text1;
  GtkWidget *dialog;
  GtkLabel *wg_label;
  GtkText *wg_text;

  dialog = create_dialog_info ();

  wg_label = GTK_LABEL (lookup_widget (dialog, "label_info"));
  gtk_label_set_text (wg_label, _("About 'Klavaro'"));

  wg_text = GTK_TEXT (lookup_widget (dialog, "text_info"));
  read_text (text, TEXT_MAXLEN, "about.txt");
  text1 = strstr (text, "(V)");
  if (text1 == NULL)
    text1 = g_strconcat ("\n ", _("Version:"), " ", PACKAGE_VERSION,
			 "\n", text, NULL);
  else
    {
      *text1 = '\0';
      text1 += 3;
      text1 = g_strconcat (text, _("Version:"), " ", PACKAGE_VERSION,
			   text1, NULL);
    }
  gtk_text_insert (wg_text, font, &black, &white, text1, -1);
  g_free (text1);

  gtk_widget_hide (window_main_);
  gtk_widget_show (dialog);
}

/**********************************************************************
 * Changing the keyboard layout
 */
void
on_button_keyboard_clicked (GtkButton * button, gpointer user_data)
{
  GtkWidget *wg;

  window_keyboard_ = create_window_keyboard ();

  wg = lookup_widget (window_main_, "entry_keyboard");
  g_free (kb_name);
  kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
  get_keyboard ();
  fill_keyboard_list (0);	/* fills all the entries */
  update_virtual_keyboard ();

  keyboard_is_modified = FALSE;

  gtk_widget_show (window_keyboard_);
  gtk_widget_hide (window_main_);
}

/**********************************************************************
 * Changing the language
 */
void
on_button_change_language_clicked (GtkButton * button, gpointer user_data)
{

  GtkWidget *dialog_confirm;

  dialog_confirm = create_dialog_confirm ();
  gtk_label_set_text (GTK_LABEL
		      (lookup_widget
		       (dialog_confirm, "label_confirm_action")), "CHANGE");
  gtk_widget_show (dialog_confirm);
}

/**********************************************************************
 * Confirm-dialog shared tasks
 */
void
on_dialog_confirm_realize (GtkWidget * widget, gpointer user_data)
{
  int nothing_to_do = 0;
  gchar *action;
  GtkLabel *wg_label;

  wg_label = GTK_LABEL (lookup_widget (widget, "label_confirm_action"));
  action = wg_label->label;
  wg_label = GTK_LABEL (lookup_widget (widget, "label_confirm"));

  if (strcmp (action, "CHANGE") == 0)
    nothing_to_do = 0;		/* Default values already applied on creation */

  else if (strcmp (action, "OVERWRITE") == 0)
    {
      gtk_window_set_title (GTK_WINDOW (widget), _("Overwrite user layout"));
      gtk_label_set_text (wg_label,
			  _
			  ("This will OVERWRITE an existent keyboard layout."));
    }

  else if (strcmp (action, "REMOVE") == 0)
    {
      gtk_window_set_title (GTK_WINDOW (widget), _("Remove user layout"));
      gtk_label_set_text (wg_label,
			  _("This will REMOVE an existent keyboard layout."));
    }

  else if (strcmp (action, "RESET") == 0)
    {
      gtk_window_set_title (GTK_WINDOW (widget), _("Reset lesson"));
      gtk_label_set_text (wg_label,
			  _("Confirming this, you will have to accomplish\n"
			    "all the lessons again, since the first one,\n"
			    "because its count will be RESET."));
    }

  else
    {
      g_warning ("No valid action for this confirm-dialog: %s", action);
      gtk_widget_destroy (widget);
    }

}

void
on_button_confirm_yes_clicked (GtkButton * button, gpointer user_data)
{
  gchar *action;
  gchar *tmp_name;
  GtkWidget *wg;

  wg = lookup_widget (GTK_WIDGET (button), "label_confirm_action");
  action = GTK_LABEL (wg)->label;

  if (strcmp (action, "CHANGE") == 0)
    change_language ();

  else if (strcmp (action, "OVERWRITE") == 0)
    {
      save_new_layout ();
      fill_keyboard_list (0);
    }

  else if (strcmp (action, "REMOVE") == 0)
    remove_user_layout ();

  else if (strcmp (action, "RESET") == 0)
    {
      cursor_blink = FALSE;
      tmp_name =
	g_strconcat (g_get_home_dir (), "/.klavaro/last_basic_lesson", NULL);
      unlink (tmp_name);
      g_free (tmp_name);
      basic_init ();
    }

  else
    g_warning ("No valid action selected for 'yes' confirm-button: %s",
	       action);

  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

void
on_button_confirm_no_clicked (GtkButton * button, gpointer user_data)
{
  gchar *action;
  GtkWidget *wg;

  wg = lookup_widget (GTK_WIDGET (button), "label_confirm_action");
  action = GTK_LABEL (wg)->label;

  if (strcmp (action, "OVERWRITE") == 0)
    {
      wg = lookup_widget (window_keyboard_, "label_kb_name");
      g_free (kb_name);
      kb_name = strdup (GTK_LABEL (wg)->label);
    }

  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}


/**********************************************************************
 * Introduction dialog
 */
#undef TEXT_MAXLEN
#define TEXT_MAXLEN 4096
void
on_button_intro_clicked (GtkButton * button, gpointer user_data)
{
  gchar text[TEXT_MAXLEN];
  GtkWidget *dialog_intro;
  GtkText *wg_text_intro;

  dialog_intro = create_dialog_info ();

  wg_text_intro = GTK_TEXT (lookup_widget (dialog_intro, "text_info"));
  read_text (text, TEXT_MAXLEN, "intro.txt");
  gtk_text_insert (wg_text_intro, font, &black, &white, text, -1);

  gtk_widget_hide (window_main_);
  gtk_widget_show (dialog_intro);
}


/**********************************************************************
 * Basic course
 */
void
on_button_basic_clicked (GtkButton * button, gpointer user_data)
{
  window_tutor_ = create_window_tutor ();
  gtk_widget_hide (window_main_);
  gtk_widget_show (window_tutor_);

  exercise_type = 1;
  basic_init ();
}

/**********************************************************************
 * Random letters training (adaptability)
 */
void
on_button_adapt_clicked (GtkButton * button, gpointer user_data)
{
  window_tutor_ = create_window_tutor ();
  gtk_widget_hide (window_main_);
  gtk_widget_show (window_tutor_);

  exercise_type = 2;
  adapt_init ();
}

/**********************************************************************
 * Random words training (velocity)
 */
void
on_button_velo_clicked (GtkButton * button, gpointer user_data)
{
  window_tutor_ = create_window_tutor ();
  gtk_widget_hide (window_main_);
  gtk_widget_show (window_tutor_);

  exercise_type = 3;
  velo_init ();
}

/**********************************************************************
 * Random texts training (fluidness)
 */
void
on_button_fluid_clicked (GtkButton * button, gpointer user_data)
{
  window_tutor_ = create_window_tutor ();
  gtk_widget_hide (window_main_);
  gtk_widget_show (window_tutor_);

  exercise_type = 4;
  fluid_init ();
}


/**********************************************************************
 * Part 2
 * Callbacks from the keyboard editor window
 */
/**********************************************************************
 * Changing the main keyboard layout
 */
void
on_entry_kb_main_insert_text (GtkEditable * editable,
			      gchar * new_text,
			      gint new_text_length,
			      gint * position, gpointer user_data)
{
  if (is_initializing_list)
    return;

  g_free (kb_name);
  kb_name = g_strndup (new_text, new_text_length);
  get_keyboard ();
  fill_keyboard_list (2);
  fill_keyboard_list (3);
  update_virtual_keyboard ();
}

/**********************************************************************
 * Changing the user keyboard layout
 */
void
on_entry_kb_user_insert_text (GtkEditable * editable,
			      gchar * new_text,
			      gint new_text_length,
			      gint * position, gpointer user_data)
{
  if (is_initializing_list)
    return;

  g_free (kb_name);
  kb_name = g_strndup (new_text, new_text_length);
  get_keyboard ();
  fill_keyboard_list (3);
  fill_keyboard_list (1);
  update_virtual_keyboard ();
}

/**********************************************************************
 * Editing a key of the keyboard
 */
void
on_entry_popup_key_changed (GtkEditable * editable, gpointer user_data)
{
  gchar *tmp_str;

  callbacks_shield = TRUE;
  tmp_str = gtk_editable_get_chars (editable, 0, -1);
  change_key (tmp_str);
  g_free (tmp_str);
  callbacks_shield = FALSE;
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (editable)));
}

void
on_button_popup_cancel_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

/**********************************************************************
 * Saving the new keyboard layout
 */
void
on_button_kb_save_clicked (GtkButton * button, gpointer user_data)
{
  gchar *tmp_path;
  FILE *fh;
  GtkWidget *wg;
  GtkWidget *dialog_confirm;

  wg = lookup_widget (GTK_WIDGET (button), "entry_kb_edit");
  kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);

  if (kb_name[0] == '\0')
    {
      g_free (kb_name);
      kb_name = g_strdup (".tmp");
    }
  else
    {
      tmp_path =
	g_strconcat (g_get_home_dir (), "/.klavaro/", kb_name, ".kbd", NULL);
      fh = fopen (tmp_path, "r");
      g_free (tmp_path);
      if (fh)
	{
	  fclose (fh);
	  dialog_confirm = create_dialog_confirm ();
	  gtk_label_set_text (GTK_LABEL (lookup_widget
					 (dialog_confirm,
					  "label_confirm_action")),
			      "OVERWRITE");
	  gtk_widget_show (dialog_confirm);
	  return;
	}
    }

  save_new_layout ();
  fill_keyboard_list (0);
}


/**********************************************************************
 * Deleting some user created layout
 */
void
on_button_kb_remove_clicked (GtkButton * button, gpointer user_data)
{
  gchar *tmp_name = NULL;
  gchar *tmp_keyb = NULL;
  GtkWidget *dialog_confirm;
  GtkWidget *wg;
  FILE *fh;

  wg = lookup_widget (window_keyboard_, "entry_kb_edit");
  tmp_keyb = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/", tmp_keyb, ".kbd", NULL);
  fh = fopen (tmp_name, "r");
  g_free (tmp_name);
  if (tmp_keyb[0] == '\0' || fh == NULL)
    {
      gdk_beep ();
      g_message ("no valid keyboard to remove...");
      g_free (tmp_keyb);
      return;
    }
  fclose (fh);
  g_free (tmp_keyb);

  dialog_confirm = create_dialog_confirm ();
  gtk_label_set_text (GTK_LABEL
		      (lookup_widget
		       (dialog_confirm, "label_confirm_action")), "REMOVE");
  gtk_widget_show (dialog_confirm);
}


/**********************************************************************
 * Selecting a keyboard layout
 */
void
on_button_keyboard_select_clicked (GtkButton * button, gpointer user_data)
{
  gint pos = 0;
  gchar *tmp_name;
  FILE *fh;
  GtkWidget *dialog;
  GtkWidget *wg;

  if (keyboard_is_modified)
    {
      wg = lookup_widget (GTK_WIDGET (button), "notebook_kb");
      gtk_notebook_set_page (GTK_NOTEBOOK (wg), 1);
      gdk_beep ();
      dialog = create_popup_message_save ();
      gtk_widget_show (dialog);
      return;
    }

  assert_user_dir ();
  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/last_keyboard", NULL);
  fh = fopen (tmp_name, "w");
  fputs (kb_name, fh);
  fclose (fh);
  g_free (tmp_name);

  wg = lookup_widget (window_main_, "entry_keyboard");
  gtk_editable_delete_text (GTK_EDITABLE (wg), 0, -1);
  gtk_editable_insert_text (GTK_EDITABLE (wg), kb_name, strlen (kb_name),
			    &pos);
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

/*
 * Closing the warning to save the created keyboard layout
 */
void
on_button_msg_save_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

/**********************************************************************
 * Allowing or not to select "load" tab on notebook
 */
void
on_notebook_kb_switch_page (GtkNotebook * notebook,
			    GtkNotebookPage * page,
			    gint page_num, gpointer user_data)
{
  if (keyboard_is_modified == FALSE)
    return;

  if (page_num == 1)
    return;

  g_timeout_add (200, &keyb_force_edit_tab, NULL);
  gdk_beep ();
}


/**********************************************************************
 * Closing the keyboard editor window
 */
void
on_button_keyboard_cancel_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

gboolean
on_window_keyboard_focus_out_event (GtkWidget * widget,
				    GdkEventFocus * event, gpointer user_data)
{
/*
  if (window_tutor_ != NULL)
    gtk_widget_destroy (widget);
*/
  return FALSE;
}

void
on_button_keyboard_close_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

gboolean
on_window_keyboard_delete_event (GtkWidget * widget,
				 GdkEvent * event, gpointer user_data)
{
  gtk_widget_destroy (widget);
  return FALSE;
}

void
on_window_keyboard_destroy (GtkObject * object, gpointer user_data)
{
  GtkWidget *wg;

  if ((window_tutor_ != NULL))
    {
      gtk_widget_show (window_tutor_);
      wg = lookup_widget (window_keyboard_, "checkbutton_hints");
      if (GTK_TOGGLE_BUTTON (wg)->active)
	{
	  window_hints_ = create_window_hints ();
	  gtk_widget_show (window_hints_);
	  wg = lookup_widget (window_tutor_, "text_tutor");
	  update_hints (gtk_editable_get_chars (GTK_EDITABLE (wg),
						cursor_pos,
						cursor_pos + 1)[0]);
	}
      return;
    }

  wg = lookup_widget (window_main_, "entry_keyboard");
  g_free (kb_name);
  kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
  get_keyboard ();
  gtk_widget_show (window_main_);
}


/**********************************************************************
 * Closing the floating hintfull hands
 */
void
on_window_hints_destroy (GtkObject * object, gpointer user_data)
{
  window_hints_ = NULL;
}

/**********************************************************************
 * Part 3
 * Callbacks from the shared tutor window
 */
/**********************************************************************
 * Selecting the basic course lessons
 */
void
on_spinbutton_lesson_changed (GtkEditable * editable, gpointer user_data)
{
  gchar *tmp_str;

  if (callbacks_shield)
    return;

  tmp_str = gtk_editable_get_chars (editable, 0, -1);
  basic_lesson = strtol (tmp_str, NULL, 0);
  g_free (tmp_str);
  get_lesson_chr_set ();
  tutor_query = QUERY_INTRO;
  basic_process_touch ('\0');
}

void
on_button_revert_lesson_clicked (GtkButton * button, gpointer user_data)
{
  basic_init ();
}

/**********************************************************************
 * Resetting the basic course
 */
void
on_button_reset_lesson_clicked (GtkButton * button, gpointer user_data)
{
  GtkWidget *dialog_confirm;

  dialog_confirm = create_dialog_confirm ();
  gtk_label_set_text (GTK_LABEL
		      (lookup_widget
		       (dialog_confirm, "label_confirm_action")), "RESET");
  gtk_widget_show (dialog_confirm);
}


/**********************************************************************
 * Other exercises popup window
 */
gint clist_other_row = -1;

void
on_button_tutor_other_clicked (GtkButton * button, gpointer user_data)
{
  popup_other_ = create_popup_other ();
  clist_other_row = -1;

  if (exercise_type == 3)
    tutor_load_clist_other (".words");
  else if (exercise_type == 4)
    tutor_load_clist_other (".paragraphs");

  callbacks_shield = TRUE;
  gtk_widget_show (popup_other_);
}

/**********************************************************************
 * Progress charts window
 */
void
on_button_tutor_stat_clicked (GtkButton * button, gpointer user_data)
{
  GtkLabel *label_stat;

  window_stat_ = create_window_stat ();
  label_stat = GTK_LABEL (lookup_widget (window_stat_, "label_stat"));

  switch (exercise_type)
    {
    case 1:
      gtk_label_set_text (label_stat, _("basic course"));
      break;
    case 2:
      gtk_label_set_text (label_stat, _("adaptability exercises"));
      break;
    case 3:
      gtk_label_set_text (label_stat, _("velocity exercises"));
      break;
    case 4:
      gtk_label_set_text (label_stat, _("fluidness exercises"));
      gtk_widget_show (lookup_widget (window_stat_, "toggle_stat_fluid"));
    }

  gtk_widget_show (window_stat_);
}

/**********************************************************************
 * Show the helping virtual keyboard
 */
void
on_button_tutor_show_clicked (GtkButton * button, gpointer user_data)
{
  if (window_hints_ != NULL)
    gtk_widget_destroy (window_hints_);

  window_keyboard_ = create_window_keyboard ();
  gtk_widget_hide (lookup_widget (window_keyboard_, "vbox_kb_left"));
  gtk_widget_show (lookup_widget (window_keyboard_, "hbox_kb_tutor"));
  gtk_widget_set_sensitive (lookup_widget (window_keyboard_, "but_space"),
			    TRUE);
  gtk_window_set_title (GTK_WINDOW (window_keyboard_), _("Virtual Keyboard"));
  gtk_label_set_text (GTK_LABEL
		      (lookup_widget (window_keyboard_, "label_virtual_kb")),
		      _("Relation between the fingers and the keys"));

  update_virtual_keyboard ();
  gtk_widget_show (window_keyboard_);
  gtk_widget_hide (window_tutor_);
}

/**********************************************************************
 * Go to the exercise's introduction
 */
void
on_button_tutor_intro_clicked (GtkButton * button, gpointer user_data)
{
  cursor_blink = FALSE;
  tutor_query = QUERY_INTRO;
  switch (exercise_type)
    {
    case 1:
      basic_update ();
      break;
    case 2:
      adapt_update ();
      break;
    case 3:
      velo_update ();
      break;
    case 4:
      fluid_update ();
      break;
    }
}

/**********************************************************************
 * Managing the sound state during exercises
 */
void
on_checkbutton_beep_toggled (GtkToggleButton * togglebutton,
			     gpointer user_data)
{
  if (callbacks_shield)
    return;

  callbacks_shield = TRUE;
  gtk_toggle_button_set_active (togglebutton, !(togglebutton->active));
  callbacks_shield = FALSE;
}


void
on_checkbutton_beep_clicked (GtkButton * button, gpointer user_data)
{
  if (callbacks_shield)
    return;

  callbacks_shield = TRUE;
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),
				!(GTK_TOGGLE_BUTTON (button)->active));
  callbacks_shield = FALSE;
}

/**********************************************************************
 * Manage the bottom messages entry on the tutor window
 */
gint entry_change_status = 0;
gchar entry_char;

void
on_entry_mesg_insert_text (GtkEditable * editable,
			   gchar * new_text,
			   gint new_text_length,
			   gint * position, gpointer user_data)
{
  if (callbacks_shield || tutor_pause)
    return;
  entry_change_status = 1;
  entry_char = new_text[0];
}

void
on_entry_mesg_changed (GtkEditable * editable, gpointer user_data)
{
  if (callbacks_shield || tutor_pause)
    return;

  if (entry_change_status != 1)
    {
      entry_change_status = 0;
      return;
    }
  entry_change_status = 0;

  switch (exercise_type)
    {
    case 1:
      basic_process_touch (entry_char);
      break;
    case 2:
      adapt_process_touch (entry_char);
      break;
    case 3:
      velo_process_touch (entry_char);
      break;
    case 4:
      fluid_process_touch (entry_char);
      break;
    }
}

void
on_entry_mesg_delete_text (GtkEditable * editable,
			   gint start_pos, gint end_pos, gpointer user_data)
{
  if (callbacks_shield || tutor_pause)
    return;

  switch (exercise_type)
    {
    case 1:
    case 2:
    case 3:
      tutor_beep ();
      break;
    case 4:
      fluid_process_touch ('\b');
      break;
    }
}

void
on_entry_mesg_activate (GtkEditable * editable, gpointer user_data)
{
  if (callbacks_shield || tutor_pause)
    return;

  callbacks_shield = TRUE;
  gtk_editable_delete_text (editable, 0, -1);
  callbacks_shield = FALSE;

  switch (exercise_type)
    {
    case 1:
      basic_process_touch ('\n');
      break;
    case 2:
      adapt_process_touch ('\n');
      break;
    case 3:
      velo_process_touch ('\n');
      break;
    case 4:
      fluid_process_touch ('\n');
      break;
    }
}

gboolean
on_entry_mesg_focus_out_event (GtkWidget * widget,
			       GdkEventFocus * event, gpointer user_data)
{
  gint pos = 0;
  gchar *tmp_str;
  GtkEditable *wg_edit;

  if (callbacks_shield || window_tutor_ == NULL)
    return FALSE;

  tutor_pause = TRUE;

  g_free (entry_mesg_buffer);
  entry_mesg_buffer = gtk_editable_get_chars (GTK_EDITABLE (widget), 0, -1);

  tmp_str =
    g_strconcat (_("PAUSED! ==>>  Press [Tab] or click here to resume."
		   "  <<== PAY ATTENTION!"), NULL);

  callbacks_shield = TRUE;
  wg_edit = GTK_EDITABLE (lookup_widget (window_tutor_, "entry_mesg"));
  gtk_editable_delete_text (wg_edit, 0, -1);
  gtk_editable_insert_text (wg_edit, tmp_str, strlen (tmp_str), &pos);
  gtk_editable_set_position (wg_edit, strlen (tmp_str));
  callbacks_shield = FALSE;

  g_free (tmp_str);
  return FALSE;
}

void
on_entry_mesg_grab_focus (GtkWidget * widget, gpointer user_data)
{
  gint pos = 0;
  GtkEditable *wg_edit;

  g_timer_start (tmr_fast);
  tutor_pause = FALSE;

  if (entry_mesg_buffer != NULL)
    {
      callbacks_shield = TRUE;
      wg_edit = GTK_EDITABLE (lookup_widget (window_tutor_, "entry_mesg"));
      gtk_editable_delete_text (wg_edit, 0, -1);
      gtk_editable_insert_text (wg_edit, entry_mesg_buffer,
				strlen (entry_mesg_buffer), &pos);
      gtk_editable_set_position (wg_edit, strlen (entry_mesg_buffer));
      callbacks_shield = FALSE;
    }
}

/**********************************************************************
 * Manage the tutor bottom message entry on focusing out of the window
 */
gboolean
on_window_tutor_focus_out_event (GtkWidget * widget,
				 GdkEventFocus * event, gpointer user_data)
{
  callbacks_shield = TRUE;
  return FALSE;
}

gboolean
on_window_tutor_focus_in_event (GtkWidget * widget,
				GdkEventFocus * event, gpointer user_data)
{
  callbacks_shield = FALSE;
  return FALSE;
}

/**********************************************************************
 * Controlling the restart action during the exercises
 */
void
on_button_tutor_restart_clicked (GtkButton * button, gpointer user_data)
{
  if (tutor_query == QUERY_END || tutor_query == QUERY_FREEZE)
    return;

  tutor_query = QUERY_INTRO;
  switch (exercise_type)
    {
    case 1:
      basic_process_touch ('\0');
      break;
    case 2:
      adapt_process_touch ('\0');
      break;
    case 3:
      velo_process_touch ('\0');
      break;
    case 4:
      fluid_process_touch ('\0');
      break;
    }

}

/**********************************************************************
 * Returning to the main menu
 */
void
on_button_tutor_close_clicked (GtkButton * button, gpointer user_data)
{
  window_tutor_ = NULL;
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

gboolean
on_window_tutor_delete_event (GtkWidget * widget,
			      GdkEvent * event, gpointer user_data)
{
  window_tutor_ = NULL;
  return FALSE;
}

void
on_window_tutor_destroy (GtkObject * object, gpointer user_data)
{
  if (window_hints_ != NULL)
    {
      gtk_widget_destroy (window_hints_);
      window_hints_ = NULL;
    }

  gtk_widget_show (window_main_);
  g_free (entry_mesg_buffer);
  entry_mesg_buffer = NULL;
}

/**********************************************************************
 * Closing the application
 */
void
on_button_tutor_exit_clicked (GtkButton * button, gpointer user_data)
{
  gtk_exit (0);
}


/**********************************************************************
 * Part 4
 * Callbacks from the "other exercises" popup window
 */
/**********************************************************************
 * Window management
 */
gboolean
on_popup_other_leave_notify_event (GtkWidget * widget,
				   GdkEventCrossing * event,
				   gpointer user_data)
{
  if (callbacks_shield)
    return FALSE;

  gtk_widget_destroy (widget);
  return FALSE;
}

gboolean
on_popup_other_enter_notify_event (GtkWidget * widget,
				   GdkEventCrossing * event,
				   gpointer user_data)
{
  if (callbacks_shield)
    {
      callbacks_shield = FALSE;
      return FALSE;
    }

  gtk_widget_destroy (widget);
  return FALSE;
}

void
on_clist_other_click_column (GtkCList * clist,
			     gint column, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (clist)));
}

/**********************************************************************
 * New file selection
 */
GtkWidget *filesel_tutor_ = NULL;

void
on_button_other_new_clicked (GtkButton * button, gpointer user_data)
{
  if (!filesel_tutor_)
    filesel_tutor_ = create_fileselection_tutor ();

  clist_other_row = -1;
  gtk_widget_destroy (popup_other_);
  gtk_widget_show (filesel_tutor_);
}

void
on_button_filesel_tutor_ok_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_hide (filesel_tutor_);

  if (exercise_type == 3)
    velo_create_dict (gtk_file_selection_get_filename
		      (GTK_FILE_SELECTION (filesel_tutor_)), TRUE);
  else if (exercise_type == 4)
    fluid_copy_text_file (gtk_file_selection_get_filename
			  (GTK_FILE_SELECTION (filesel_tutor_)));
}

void
on_button_filesel_tutor_cancel_clicked (GtkButton * button,
					gpointer user_data)
{
  gtk_widget_hide (filesel_tutor_);
}

void
on_fileselection_tutor_destroy (GtkObject * object, gpointer user_data)
{
  filesel_tutor_ = NULL;
}

/**********************************************************************
 * Remove some user exercise
 */
void
on_button_other_remove_clicked (GtkButton * button, gpointer user_data)
{
  gchar *text[1];
  GtkCList *clist;

  if (clist_other_row < 0)
    return;

  clist = GTK_CLIST (lookup_widget (GTK_WIDGET (button), "clist_other"));

  gtk_clist_get_text (clist, clist_other_row, 0, text);
  if (!strcmp (text[0], _("--> Default")))
    {
      gdk_beep ();
      return;
    }

  *text = g_strconcat (g_get_home_dir (), "/.klavaro/", text[0], NULL);
  if (exercise_type == 3)
    *text = g_strconcat (text[0], ".words", NULL);
  else if (exercise_type == 4)
    *text = g_strconcat (text[0], ".paragraphs", NULL);
  unlink (text[0]);
  g_free (text[0]);

  gtk_clist_remove (clist, clist_other_row);
}

/**********************************************************************
 * Artificially tracking the selected row. How to do this by other way?
 */
void
on_clist_other_select_row (GtkCList * clist,
			   gint row,
			   gint column, GdkEvent * event, gpointer user_data)
{
  clist_other_row = row;
}

void
on_clist_other_unselect_row (GtkCList * clist,
			     gint row,
			     gint column,
			     GdkEvent * event, gpointer user_data)
{
  clist_other_row = -1;
}

/**********************************************************************
 * Closing the exercise selection popup
 */
void
on_clist_other_destroy (GtkObject * object, gpointer user_data)
{
  gchar *text[1];

  if (clist_other_row < 0)
    return;

  gtk_clist_get_text (GTK_CLIST (object), clist_other_row, 0, text);

  if (exercise_type == 3)
    {
      init_word_list (text[0]);
      tutor_query = QUERY_INTRO;
      velo_process_touch ('\0');
    }
  else if (exercise_type == 4)
    {
      init_paragraph_list (text[0]);
      tutor_query = QUERY_INTRO;
      fluid_process_touch ('\0');
    }
}

/**********************************************************************
 * Part 5
 * Callbacks from the shared progress window
 */
/**********************************************************************
 * Initialization of the chart
 */
void
on_frame_stat_realize (GtkWidget * widget, gpointer user_data)
{
  initialize_chart (widget);
}

/**********************************************************************
 * Show accuracy chart
 */
void
on_toggle_stat_accur_toggled (GtkToggleButton * togglebutton,
			      gpointer user_data)
{
  GtkWidget *wg;

  wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_velo");
  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
    {
      wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_fluid");
      if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
	{
	  if (!gtk_toggle_button_get_active (togglebutton))
	    gtk_toggle_button_set_active (togglebutton, TRUE);
	  return;
	}
      else if (!gtk_toggle_button_get_active (togglebutton))
	return;
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);
    }
  else if (!gtk_toggle_button_get_active (togglebutton))
    return;
  else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);

  plot_stat_log (1);
}

/**********************************************************************
 * Show velocity chart
 */
void
on_toggle_stat_velo_toggled (GtkToggleButton * togglebutton,
			     gpointer user_data)
{
  GtkWidget *wg;

  wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_accur");
  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
    {
      wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_fluid");
      if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
	{
	  if (!gtk_toggle_button_get_active (togglebutton))
	    gtk_toggle_button_set_active (togglebutton, TRUE);
	  return;
	}
      else if (!gtk_toggle_button_get_active (togglebutton))
	return;
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);
    }
  else if (!gtk_toggle_button_get_active (togglebutton))
    return;
  else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);

  plot_stat_log (2);
}

/**********************************************************************
 * Show fluidness chart
 */
void
on_toggle_stat_fluid_toggled (GtkToggleButton * togglebutton,
			      gpointer user_data)
{
  GtkWidget *wg;

  wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_accur");
  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
    {
      wg = lookup_widget (GTK_WIDGET (togglebutton), "toggle_stat_velo");
      if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (wg)))
	{
	  if (!gtk_toggle_button_get_active (togglebutton))
	    gtk_toggle_button_set_active (togglebutton, TRUE);
	  return;
	}
      else if (!gtk_toggle_button_get_active (togglebutton))
	return;
      else
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);
    }
  else if (!gtk_toggle_button_get_active (togglebutton))
    return;
  else
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (wg), FALSE);

  plot_stat_log (3);
}

/**********************************************************************
 * Closing the progress window
 */

void
on_button_stat_close_clicked (GtkButton * button, gpointer user_data)
{
  gtk_widget_destroy (gtk_widget_get_toplevel (GTK_WIDGET (button)));
}

gboolean
on_window_stat_delete_event (GtkWidget * widget,
			     GdkEvent * event, gpointer user_data)
{
  gtk_widget_destroy (widget);
  return FALSE;
}

/**********************************************************************
 * Last part
 * Editing the keyboard. Because it's so large, it's here at the end.
 */
/**********************************************************************
 * Shift keys
 */
void
on_toggle_shift1_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
  gboolean tog_state;
  GtkWidget *tog;

  tog = lookup_widget (GTK_WIDGET (togglebutton), "toggle_shift2");
  tog_state = gtk_toggle_button_get_active (togglebutton);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (tog), tog_state);
  update_virtual_keyboard ();

  if (window_tutor_ != NULL)
    edit_key (GTK_BUTTON (togglebutton), "hands_0.xpm");
}

void
on_toggle_shift2_toggled (GtkToggleButton * togglebutton, gpointer user_data)
{
  gboolean tog_state;
  GtkWidget *tog;

  tog = lookup_widget (GTK_WIDGET (togglebutton), "toggle_shift1");
  tog_state = gtk_toggle_button_get_active (togglebutton);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (tog), tog_state);
  update_virtual_keyboard ();

  if (window_tutor_ != NULL)
    edit_key (GTK_BUTTON (togglebutton), "hands_0.xpm");
}

/**********************************************************************
 * Space bar
 */
void
on_but_space_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_5.xpm");
}

/**********************************************************************
 * All the other keys
 */
void
on_but1_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but2_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but3_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_2.xpm");
}

void
on_but4_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_3.xpm");
}

void
on_but5_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but6_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but7_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but8_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but9_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_7.xpm");
}

void
on_but10_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_8.xpm");
}

void
on_but11_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but12_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but13_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but14_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but_1_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but_2_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_2.xpm");
}

void
on_but_3_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_3.xpm");
}

void
on_but_4_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but_5_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but_6_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but_7_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but_8_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_7.xpm");
}

void
on_but_9_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_8.xpm");
}

void
on_but_10_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but_11_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but_12_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but__1_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but__2_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_2.xpm");
}

void
on_but__3_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_3.xpm");
}

void
on_but__4_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but__5_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but__6_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but__7_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but__8_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_7.xpm");
}

void
on_but__9_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_8.xpm");
}

void
on_but__10_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but__11_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but__12_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but___1_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but___2_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_1.xpm");
}

void
on_but___3_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_2.xpm");
}

void
on_but___4_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_3.xpm");
}

void
on_but___5_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but___6_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_4.xpm");
}

void
on_but___7_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but___8_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_6.xpm");
}

void
on_but___9_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_7.xpm");
}

void
on_but___10_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_8.xpm");
}

void
on_but___11_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

void
on_but___12_clicked (GtkButton * button, gpointer user_data)
{
  edit_key (button, "hands_9.xpm");
}

/**********************************************************************
 * The end
 */
