#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <libgen.h>
#include <unistd.h>

#include "support.h"

#include "global.h"
#include "callbacks.h"
#include "translation.h"
#include "keyboard.h"
#include "velocity.h"
#include "fluidness.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_tutor_;
extern gchar *data_path;
extern gboolean cursor_blink;
extern gboolean cursor_is_initializing;
extern gint tutor_query;
extern gint cursor_pos;
extern gint n_touchs;
extern gint n_errors;
extern gint retro_pos;
extern gint correcting;
extern gdouble elapsed_time;
extern gdouble touch_time[2001];
extern GTimer *tmr;
extern GTimer *tmr_fast;
extern GdkFont *font;
extern GdkColor black;
extern GdkColor white;
extern Paragraph par;

/**********************************************************************
 * Global variables used only here
 */
gchar par_list_name[21];

/**********************************************************************
 * Reads paragraphs from the text file.
 */
void
init_paragraph_list (gchar * list_name)
{
  gint len;
  gchar *tmp_name;
  gchar tmp_code[16];
  gchar str_1000[1001];
  FILE *fh;

  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/last_paragraph_list", NULL);
  if (list_name && strcmp (list_name, _("--> Default")))
    {
      assert_user_dir ();
      if ((fh = fopen (tmp_name, "w")))
	{
	  fprintf (fh, "%s", list_name);
	  fclose (fh);
	}
      else
	g_message ("could not write the file:\n %s", tmp_name);

      g_free (tmp_name);
      tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/",
			      list_name, ".paragraphs", NULL);
      g_message ("loading text file: %s.paragraphs", list_name);
      strncpy (par_list_name, list_name, 20);
      par_list_name[20] = '\0';
    }
  else
    {
      unlink (tmp_name);
      g_free (tmp_name);
      get_lang_env (tmp_code);
      tmp_name = g_strconcat (data_path, tmp_code, ".paragraphs", NULL);
      g_message ("loading text file: %s.paragraphs", tmp_code);
      strcpy (par_list_name, "Default");
    }

  if ((fh = fopen (tmp_name, "r")))
    {
      g_free (par.str);
      par.str = g_strconcat ("", NULL);
      par.n = 0;
      while (fgets (str_1000, 1001, fh))
	{
	  len = strlen (str_1000);
	  if (len < 3)
	    continue;
	  if (!ispunct (str_1000[len - 2]))
	    str_1000[len - 2] = '.';
	  if (str_1000[len - 1] != '\n')
	    str_1000[len - 1] = '\n';
	  par.str = g_strconcat (par.str, str_1000, NULL);
	  par.n++;
	}
      fclose (fh);
      g_message ("text file loaded!\n");
    }
  else
    {
      g_message ("could not open the file: %s", tmp_name);
      g_free (tmp_name);
      get_lang_env (tmp_code);
      if (strcmp (tmp_code, "C") == 0)
	g_error ("so, we must quit!");
      setenv ("LANGUAGE", "C", 1);
      init_paragraph_list (list_name);
      setenv ("LANGUAGE", tmp_code, 1);
      return;
    }
  g_free (tmp_name);
}

/**********************************************************************
 * Get from the structure 'par' the paragraph defined by 'index'
 */
gchar *
get_par (gint index)
{
  gint i;
  gint size;
  gchar *par_1;
  gchar *par_2;
  gchar *par_i;

  par_1 = par.str;
  par_2 = strchr (par.str, '\n') + 1;
  for (i = 0; i < index && i < par.n; i++)
    {
      par_1 = par_2;
      par_2 = strchr (par_2, '\n') + 1;
    }
  size = par_2 - par_1;
  if (size < 1)
    {
      g_warning ("internal error while picking the paragraph %i.", index);
      return (NULL);
    }
  par_i = g_malloc (size + 1);
  memmove (par_i, par_1, size);
  par_i[size] = '\0';

  return (par_i);
}

/**********************************************************************
 * Initialize the fluid exercise window.
 */
void
fluid_init ()
{
  gchar *tmp_name;
  gchar pl_name[1000];
  FILE *fh;

  gtk_window_set_title (GTK_WINDOW (window_tutor_), _("Klavaro - Fluidness"));
  gtk_widget_hide (lookup_widget (window_tutor_, "frame_lesson"));
  gtk_widget_hide (lookup_widget (window_tutor_, "vseparator_tutor"));
  gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_show"));
  gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_other"));

  gtk_label_set_text (GTK_LABEL
		      (lookup_widget (window_tutor_, "label_heading")),
		      _
		      ("Fluidness exercises: accuracy typing good sense paragraphs."));

  if (par.n == 0)
    {
      tmp_name =
	g_strconcat (g_get_home_dir (), "/.klavaro/last_paragraph_list",
		     NULL);
      if ((fh = fopen (tmp_name, "r")))
	{
	  fscanf (fh, "%s", pl_name);
	  g_free (tmp_name);
	  tmp_name =
	    g_strconcat (g_get_home_dir (), "/.klavaro/", pl_name,
			 ".paragraphs", NULL);
	  fclose (fh);
	  if ((fh = fopen (tmp_name, "r")))
	    {
	      fclose (fh);
	      init_paragraph_list (pl_name);
	    }
	}
      g_free (tmp_name);
    }
  if (par.n == 0)
    init_paragraph_list (NULL);

  tutor_query = QUERY_INTRO;
  fluid_update ();
}

/**********************************************************************
 * Update what is shown in the tutor window. 
 */
void
fluid_update ()
{
  gchar text[4096];
  GtkText *wg_text;
  GtkEditable *wg_edit;
  GtkScrolledWindow *wg_scroll;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  wg_edit = GTK_EDITABLE (lookup_widget (window_tutor_, "text_tutor"));
  wg_scroll =
    GTK_SCROLLED_WINDOW (lookup_widget
			 (window_tutor_, "scrolledwindow_tutor_main"));

  gtk_text_freeze (wg_text);
  switch (tutor_query)
    {

    case QUERY_INTRO:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      read_text (text, 4096, "fluid_intro.txt");
      gtk_text_insert (wg_text, font, &black, &white, text, -1);
      tutor_message (_("Press any key to start the exercise. "));
      gtk_scrolled_window_set_policy (wg_scroll, GTK_POLICY_NEVER,
				      GTK_POLICY_ALWAYS);
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_START:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      draw_random_paragraphs ();
      tutor_message (_("Start typing when you are ready. "));
      gtk_scrolled_window_set_policy (wg_scroll, GTK_POLICY_NEVER,
				      GTK_POLICY_NEVER);
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_PROCESS_TOUCHS:
      tutor_message (_("STARTED! => "));
      gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_END:
      tutor_message (_("<<==  End of exercise."
		       " Press [Enter] to start another. ==>> "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;
    }
  gtk_text_thaw (wg_text);
}

/**********************************************************************
 * Responds to each touch of the user, according to the 'tutor_query' mode
 */
void
fluid_process_touch (gchar user_chr)
{
  gulong micros;
  GtkText *wg_text;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  switch (tutor_query)
    {
    case QUERY_PROCESS_TOUCHS:
      break;

    case QUERY_INTRO:
      tutor_query = QUERY_START;
      fluid_update ();
      n_touchs = 0;
      n_errors = 0;
      retro_pos = 0;
      correcting = 0;
      gtk_text_set_point (wg_text, 2);
      cursor_pos = 2;
      if (!cursor_is_initializing)
	{
	  cursor_blink = FALSE;
	  cursor_is_initializing = TRUE;
	  g_timeout_add (TIME_CURSOR_ON + 100, &init_cursor, NULL);
	}
      return;

    case QUERY_START:
      tutor_query = QUERY_PROCESS_TOUCHS;
      fluid_update ();
      touch_time[0] = 0;
      g_timer_start (tmr_fast);
      g_timer_start (tmr);
      forward_backward_cursor (user_chr);
      return;

    case QUERY_END:
      if (user_chr == '\n')
	{
	  tutor_query = QUERY_INTRO;
	  fluid_process_touch ('\0');
	}
      else
	{
	  tutor_beep ();
	  fluid_update ();
	}
      return;

    default:
      g_warning ("fluid_process_touch(): the program must not pass here!");
      return;
    }

  /*
   * It is time to analise the correctness of typing.
   */
  forward_backward_cursor (user_chr);
  if (cursor_pos >= gtk_text_get_length (wg_text))
    {
      elapsed_time = g_timer_elapsed (tmr, &micros);
      elapsed_time += rint (micros / 1.e6);
      n_touchs += n_errors;
      fluid_end_exercise ();
      tutor_query = QUERY_END;
      fluid_update ();
      tutor_beep ();
      return;
    }
}

/**********************************************************************
 * Put on the screen the final results
 */
#define MAX_TEXT_LEN 1024
void
fluid_end_exercise ()
{
  gchar *tmp_locale;
  gchar *str;
  gchar tmp_str[MAX_TEXT_LEN];
  gchar *tmp_name;
  GtkText *wg_text;
  FILE *fh;
  time_t tmp_time;
  struct tm *ltime;
  gint minutes;
  gint seconds;
  gdouble accuracy;
  gdouble touchs_per_second;
  gdouble words_per_minute;
  gdouble fluidness;
  gdouble standard_deviation;
  gdouble sum, average;
  gdouble sample;
  gint i;

  /*
   * Calculate statistics
   */
  minutes = ((gulong) elapsed_time) / 60;
  seconds = ((gulong) elapsed_time) % 60;

  accuracy = 100 * (1.0 - (gfloat) n_errors / n_touchs);
  touchs_per_second = (n_touchs - n_errors) / elapsed_time;
  words_per_minute = 10 * touchs_per_second;

  /*
   * "Magic" fluidness calculation
   */
  sum = 0;
  for (i = 2; i < (gint) touch_time[0]; i++)
    {
      sample = sqrt (1 / touch_time[i]);
      sum += sample;
    }
  average = sum / (i - 2);

  sum = 0;
  for (i = 2; i < (gint) touch_time[0]; i++)
    {
      sample = sqrt (1 / touch_time[i]);
      sum += (sample - average) * (sample - average);
    }
  standard_deviation = sqrt (sum / (i - 3));
  fluidness = 100 * (1 - standard_deviation / average);
  if (fluidness < 0)
    fluidness = 0;

  /*
   * Changing to "C" locale
   */
  tmp_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
  setlocale (LC_NUMERIC, "C");

  /*
   * Logging
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/stat_fluid.txt", NULL);
  assert_user_dir ();
  if (!(fh = fopen (tmp_name, "r")))
    {
      fh = fopen (tmp_name, "w");
      fprintf (fh, "Accuracy\tVelocity\tFluidness\tDate\tHour\tLesson\n");
    }
  else
    {
      fclose (fh);
      fh = fopen (tmp_name, "a");
    }
  if (fh)
    {
      tmp_time = time (NULL);
      ltime = localtime (&tmp_time);
      fprintf (fh, "%.2f\t%.2f\t%.2f\t%i-%2.2i-%2.2i\t%2.2i:%2.2i\t%s\n",
	       accuracy, words_per_minute, fluidness, (ltime->tm_year) + 1900,
	       (ltime->tm_mon) + 1, (ltime->tm_mday), (ltime->tm_hour),
	       (ltime->tm_min), par_list_name);
      fclose (fh);
    }
  else
    g_message ("not able to log on this file:\n %s", tmp_name);

  /*
   * Logging the fluidness results of the last session
   */
  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/deviation_fluid.txt", NULL);
  if ((fh = fopen (tmp_name, "w")))
    {
      g_message ("writing further fluidness results at:\n %s", tmp_name);
      fprintf (fh, "Samples\t\"Average:\"\t%g\t\"Std. dev.:\"\t%g\n",
	       average, standard_deviation);
      for (i = 2; i < (gint) touch_time[0]; i++)
	fprintf (fh, "%g\n", sqrt (1 / touch_time[i]));
      fclose (fh);
    }
  else
    g_message ("not able to log on this file:\n %s", tmp_name);
  g_free (tmp_name);

  /*
   * Coming back to the right locale
   */
  setlocale (LC_NUMERIC, tmp_locale);
  g_free (tmp_locale);

  /*
   * Print statistics
   */
  str = g_strconcat ("\n\n ", _("STATISTICS"), "\n\n ",
		     _("Elapsed time:"), " %i ",
		     _("minute(s) and"), " %i ", _("seconds"), "\n\n ",
		     _("Accuracy:"), " %.1f %%\t\t",
		     _("Goal:"), " %i %%", "\n\n ",
		     _("Characters per minute:"), " %.2f\t",
		     _("Goal:"), " %i ", _("(CPS)"), "\n ",
		     _("Words per minute:"), " %.1f\t",
		     _("Goal:"), " %i ", _("(WPM)"), "\n\n ",
		     _("Fluidness:"), " %.1f %%\t\t",
		     _("Goal:"), " %i %%", "\n\n", NULL);
  sprintf (tmp_str, str, minutes, seconds, accuracy, 98,
	   words_per_minute / 10, 5, words_per_minute, 50, fluidness, 90);
  g_free (str);

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  gtk_text_insert (wg_text, font, &black, &white, tmp_str, -1);

  /*
   * Comments
   */
  strcpy (tmp_str, _(" Comments:\n\n"));
  str = tmp_str + strlen (tmp_str);
  strcpy (str, "...");
  if (accuracy < 98)
    sprintf (str,
	     _(" Your accuracy rate is too low.\n"
	       " I want it greater than 98 %%.\n"));
  else if (words_per_minute < 50)
    sprintf (str, _(" You type accurately but not fast.\n"
		    " Can you reach 50 WPM?\n"));
  else if (fluidness < 70)
    sprintf (str,
	     _(" Your rhythm is not so constant. Calm down.\n"
	       " For now, try to make the fluidness greater than 70 %%.\n"));
  else if (fluidness < 90)
    sprintf (str,
	     _(" You are almost getting there. Type more fluently.\n"
	       " I want a fluidness greater than 90 %%.\n"));
  else if (words_per_minute < 80)
    sprintf (str,
	     _(" Congratulations!\n"
	       " It seems to me that you are a professional.\n"
	       " You don't need this program (me) anymore.\n"
	       " Hope you have enjoyed. Thanks and be happy!\n"));
  else
    sprintf (str,
	     _(" How can you type so fast?\n"
	       " You have exceeded all my expectations.\n"
	       " Are you a machine? Could you teach me?\n"
	       " I can not help you anymore. Go to an expert!\n"));

  gtk_text_insert (wg_text, NULL, &black, &white, tmp_str, -1);
}


/**********************************************************************
 * Draw random sentences selected from a '.paragraphs' file
 */
void
draw_random_paragraphs ()
{
  gint i, j;
  gint rand_i[3];
  gchar *text;

  for (i = 0; (i < 3) && (i < par.n); i++)
    {
      do
	{
	  rand_i[i] = rand () % par.n;
	  for (j = 0; j < i; j++)
	    {
	      if (rand_i[i] == rand_i[j])
		rand_i[i] = par.n;
	    }
	}
      while (rand_i[i] == par.n);

      text = get_par (rand_i[i]);
      draw_paragraph (text);
      g_free (text);
    }
}

/**********************************************************************
 * Copy the file 'file_name' so that it can be used as a customized
 * exercise.
 */
void
fluid_copy_text_file (gchar * file_name)
{
  gint tmp_chr;
  gchar *pars_path;
  gchar *pars_name;
  FILE *fh_source;
  FILE *fh_destiny;

  if (!file_name)
    {
      gdk_beep ();
      g_warning ("fluid_copy_text_file(): null file name as argument.");
      return;
    }

  if (!(fh_source = fopen (file_name, "r")))
    {
      gdk_beep ();
      g_warning ("couldn't open the file:\n %s\n", file_name);
      return;
    }

  pars_path = g_strconcat (g_get_home_dir (), "/.klavaro/", NULL);
  pars_path = g_strconcat (pars_path, basename (file_name), NULL);
  if (strrchr (strrchr (pars_path, '/'), '.'))
    *(strrchr (pars_path, '.')) = '\0';
  pars_name = g_strconcat (basename (pars_path), NULL);
  pars_path = g_strconcat (pars_path, ".paragraphs", NULL);

  assert_user_dir ();
  if (!(fh_destiny = fopen (pars_path, "w")))
    {
      gdk_beep ();
      g_warning ("couldn't create the file:\n %s\n", pars_path);
      fclose (fh_source);
      g_free (pars_path);
      g_free (pars_name);
      return;
    }
  g_free (pars_path);

  while ((tmp_chr = fgetc (fh_source)) != EOF)
    fputc (tmp_chr, fh_destiny);
  fclose (fh_source);
  fclose (fh_destiny);

  init_paragraph_list (pars_name);
  tutor_query = QUERY_INTRO;
  fluid_process_touch ('\0');

  velo_create_dict (file_name, FALSE);
  g_free (pars_name);
}
