/*
 * Some auxiliar functions shared by other functions
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>
#include <sys/stat.h>
#include <gtkextra/gtkextra.h>

#include "support.h"
#include "interface.h"

#include "global.h"
#include "translation.h"
#include "keyboard.h"

/**********************************************************************
 * Wide global variables
 */
GtkWidget *window_main_ = NULL;
GtkWidget *window_hints_ = NULL;
gchar *data_path = NULL;
gboolean callbacks_shield = FALSE;
GdkFont *font;
GdkColor black;
GdkColor white;
GdkColor red;
GdkColor green;
GdkColor blue;
GdkColor darkgreen;

/****************************************************************************
 * Global variables for the keyboard definition and edition.
 */
GtkWidget *window_keyboard_;
gboolean is_initializing_list = TRUE;
gboolean keyboard_is_modified = FALSE;
gchar *kb_name = NULL;
gchar lowerchars[4][KEY_LINE_LEN + 1];
gchar upperchars[4][KEY_LINE_LEN + 1];
gchar *consonants = NULL;
gchar *symbols = NULL;

/****************************************
 * Basic course global variables
 */
gint basic_lesson;

/**********************************************************************
 * Velocity exercise global variables
 */
GList *word_list = NULL;
gint word_list_len = 0;

/**********************************************************************
 * Fluidness exercise global variables
 */
Paragraph par = { NULL, 0 };

/**********************************************************************
 * Shared tutor window global variables
 */
GtkWidget *window_tutor_ = NULL;
GtkWidget *popup_other_ = NULL;
GtkWidget *window_stat_;
GTimer *tmr;
GTimer *tmr_fast;
gdouble elapsed_time;
gdouble touch_time[2001];
gint n_touchs;
gint n_errors;
gint cursor_pos;
gint retro_pos;
gint correcting;
gint exercise_type = 0;
gint tutor_query = 0;
gboolean tutor_pause = FALSE;
gboolean cursor_blink = FALSE;
gboolean cursor_is_initializing = FALSE;
gchar *entry_mesg_buffer = NULL;


/**********************************************************************
 * Initialize the value of the global variables
 */
void
initialize_global_variables ()
{
  gint pos = 0;
  gchar *tmp_name = NULL;
  gchar tmp_str[21];
  FILE *fh;
  GtkWidget *wg;

  /*
   * Get a valid data path.
   * First searches the path at the source directory.
   */
  data_path = g_strconcat ("../data/", NULL);
  tmp_name = g_strconcat (data_path, "basic_lessons.txt", NULL);
  fh = fopen (tmp_name, "r");
  if (fh == NULL)
    {
      g_free (data_path);
      g_free (tmp_name);
      data_path = g_strconcat (PACKAGE_DATA_DIR, "/", NULL);
      tmp_name = g_strconcat (data_path, "basic_lessons.txt", NULL);
      fh = fopen (tmp_name, "r");
    }
  if (fh == NULL)
    {
      g_free (tmp_name);
      g_error ("couldn't find the file at data path: basic_lessons.txt");
    }
  fclose (fh);
  g_free (tmp_name);

  /*
   * Set the initial keyboard to use
   */
  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/last_keyboard", NULL);
  wg = lookup_widget (window_main_, "entry_keyboard");
  if (!(fh = fopen (tmp_name, "r")))
    gtk_editable_insert_text (GTK_EDITABLE (wg), DEFAULT_KEYBOARD,
			      strlen (DEFAULT_KEYBOARD), &pos);
  else
    {
      fgets (tmp_str, 21, fh);
      if (tmp_str[strlen (tmp_str) - 1] == '\n')
	tmp_str[strlen (tmp_str) - 1] = '\0';
      fclose (fh);
      gtk_editable_insert_text (GTK_EDITABLE (wg), tmp_str, strlen (tmp_str),
				&pos);
    }
  g_free (tmp_name);

  kb_name = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);
  get_keyboard ();

  /*
   * Initialize the available languages at main menu
   */
  set_combo_language ();

  /*
   * Set the font to use at the tutor window
   */
  font = gdk_font_load ("-*-*-bold-r-*-*-*-140-*-*-m-*-*-*");
  if (!font)
    g_warning
      ("main window: no adequate font loaded (bold, roman, 140 pt, monospaced)");

  /*
   * Set the magic colors
   */
  black.pixel = 0;
  black.red = 0x0000;
  black.green = 0x0000;
  black.blue = 0x0000;

  white.pixel = 1;
  white.red = 0xFFFF;
  white.green = 0xFFFF;
  white.blue = 0xFFFF;

  red.pixel = 2;
  red.red = 0xF0FF;
  red.green = 0xC000;
  red.blue = 0xC400;

  green.pixel = 3;
  green.red = 0xA000;
  green.green = 0xF000;
  green.blue = 0xC000;

  blue.pixel = 4;
  blue.red = 0xE400;
  blue.green = 0xEA00;
  blue.blue = 0xF0FF;

  darkgreen.pixel = 5;
  darkgreen.red = 0xB000;
  darkgreen.green = 0xD000;
  darkgreen.blue = 0xA000;

  /*
   * Randomize
   */
  srand (time (0));

  /*
   * Initialize timers
   */
  tmr = g_timer_new ();
  tmr_fast = g_timer_new ();
}

/**********************************************************************
 * Get the code of the current language set by the environmental variables
 * LANGUAGE, LC_ALL, LC_MESSAGES or LANG.
 */
void
get_lang_env (gchar * tmp_code)
{
  gchar *tmp_str = NULL;

  tmp_str = getenv ("LANGUAGE");

  if (tmp_str == NULL)
    tmp_str = getenv ("LC_ALL");
  else if (strlen (tmp_str) < 1)
    tmp_str = getenv ("LC_ALL");

  if (tmp_str == NULL)
    tmp_str = getenv ("LC_MESSAGES");
  else if (strlen (tmp_str) < 1)
    tmp_str = getenv ("LC_MESSAGES");

  if (tmp_str == NULL)
    tmp_str = getenv ("LANG");
  else if (strlen (tmp_str) < 1)
    tmp_str = getenv ("LANG");

  if (tmp_str == NULL)
    strcpy (tmp_code, "C");
  else if (strlen (tmp_str) < 1)
    strcpy (tmp_code, "C");
  else
    strcpy (tmp_code, tmp_str);
}

/**********************************************************************
* Search for the user directory and create it if not found
*/
void
assert_user_dir ()
{
  gchar *tmp_name;
  FILE *fh;

  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro", NULL);
  fh = fopen (tmp_name, "r");
  if (fh == NULL)
    {
      g_message ("creating an empty user folder: ~/.klavaro");
      mkdir (tmp_name, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
      fh = fopen (tmp_name, "r");
      if (fh == NULL)
	g_error ("could not creat it, so we must quit!");
    }
  g_free (tmp_name);
  fclose (fh);
}

/**********************************************************************
 * Beeps (or not) at the user, in the tutor window
 */
void
tutor_beep ()
{
  GtkWidget *wg;

  wg = lookup_widget (window_tutor_, "checkbutton_beep");
  if ((GTK_TOGGLE_BUTTON (wg))->active)
    gdk_beep ();
}

/**********************************************************************
 * Put 'mesg' in the message entry line of the shared tutor window
 */
void
tutor_message (gchar * mesg)
{
  /*
     GtkEditable *wg;
     gint pos = 0;
   */
  gchar *tmp_str = NULL;
  GtkEntry *wg;

  if (!mesg)
    {
      g_message ("tutor_message() --> can't show NULL message!");
      return;
    }

  tmp_str = g_strconcat (" ", mesg, NULL);

  if (tutor_pause)
    {
      g_free (entry_mesg_buffer);
      entry_mesg_buffer = g_strdup (tmp_str);
    }
  else
    {
      callbacks_shield = TRUE;
      wg = GTK_ENTRY (lookup_widget (window_tutor_, "entry_mesg"));
      gtk_entry_set_text (wg, tmp_str);
      /*
         wg = GTK_EDITABLE (lookup_widget (window_tutor_, "entry_mesg"));
         gtk_editable_delete_text (wg, 0, -1);
         gtk_editable_insert_text (wg, tmp_str, strlen (tmp_str), &pos);
         gtk_editable_set_position (wg, strlen (tmp_str) - 1);
       */
      callbacks_shield = FALSE;

    }

  g_free (tmp_str);
}

/**********************************************************************
 * Formats and draws one paragraph at the tutor window
 */
void
draw_paragraph (gchar * text)
{
  gint column;
  gchar *next_space;
  GtkText *text_wg;

  text_wg = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  text[strlen (text) - 1] = 182;

  gtk_text_insert (text_wg, font, &black, &white, "\n ", 2);

  column = 1;
  while (1)
    {
      next_space = strchr (text, ' ');
      if (!next_space)
	{
	  if (column + strlen (text) > 54)
	    gtk_text_insert (text_wg, font, &black, &white, "\n ", 2);
	  break;
	}

      column += (next_space - text) + 1;
      if (column > 55)
	{
	  column = (next_space - text) + 2;
	  gtk_text_insert (text_wg, font, &black, &white, "\n ", 2);
	}

      *next_space = '\0';
      gtk_text_insert (text_wg, font, &black, &blue, text, -1);
      gtk_text_insert (text_wg, font, &black, &blue, " ", 1);
      text = next_space + 1;
    }

  gtk_text_insert (text_wg, font, &black, &blue, text, -1);
  gtk_text_insert (text_wg, font, &black, &white, "\n", 1);
}

/**********************************************************************
 * Load the list of files to include in the set of "other exercises"
 */
void
tutor_load_clist_other (gchar * file_name_end)
{
  gchar *tmp_str;
  gchar *tmp_name[1];
  GtkCList *clist;
  DIR *dir;
  struct dirent *dentry;

  assert_user_dir ();
  tmp_name[0] = g_strconcat (g_get_home_dir (), "/.klavaro/", NULL);
  dir = opendir (tmp_name[0]);
  g_free (tmp_name[0]);

  clist =
    GTK_CLIST (lookup_widget (GTK_WIDGET (popup_other_), "clist_other"));
  while ((dentry = readdir (dir)))
    {
      if (strlen (dentry->d_name) < 5)
	continue;
      if (!(tmp_str = strrchr (dentry->d_name, '.')))
	continue;
      if (strcmp (file_name_end, tmp_str))
	continue;

      tmp_name[0] = g_strconcat (dentry->d_name, NULL);
      *(strrchr (tmp_name[0], '.')) = '\0';
      gtk_clist_append (clist, tmp_name);
      g_free (tmp_name[0]);
    }
  closedir (dir);

  tmp_name[0] = g_strconcat (_("--> Default"), NULL);
  gtk_clist_append (clist, tmp_name);
  g_free (tmp_name[0]);
}

/**********************************************************************
 * Advances the cursor one position and test for correctness,
 * in the shared tutor window.
 * Updates the variables:
 * cursor_pos, n_touchs and n_errors.
 */
void
forward_cursor (gchar user_chr)
{
  gchar real_chr;
  gchar *chr_str;
  GtkText *text_wg;
  GtkEditable *edit_wg;

  n_touchs++;

  text_wg = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (text_wg);

  chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
  real_chr = chr_str[0];

  gtk_text_forward_delete (text_wg, 1);

  if (user_chr == real_chr || ((real_chr == (gchar) 182) && user_chr == '\n'))
    gtk_text_insert (text_wg, font, &black, &green, chr_str, 1);
  else
    {
      tutor_beep ();
      gtk_text_insert (text_wg, font, &black, &red, chr_str, 1);
      n_errors++;
    }
  g_free (chr_str);

  cursor_blink = FALSE;
  cursor_pos++;

  if (real_chr == (gchar) 182)
    {
      cursor_pos += 3;
      if (cursor_pos < gtk_text_get_length (text_wg) - 3)
	gtk_text_set_point (text_wg, cursor_pos);
      else
	return;
    }

  chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
  if (chr_str[0] == '\n')
    {
      gtk_text_freeze (text_wg);
      cursor_pos += 2;
      gtk_text_set_point (text_wg, cursor_pos);
      gtk_text_thaw (text_wg);
    }
  g_free (chr_str);

  cursor_blink = TRUE;

  if (real_chr == (gchar) 182)
    gtk_text_freeze (text_wg);

  chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
  update_hints (chr_str[0]);
  gtk_text_forward_delete (text_wg, 1);
  gtk_text_insert (text_wg, font, &white, &black, chr_str, 1);
  gtk_text_set_point (text_wg, gtk_text_get_point (text_wg) - 1);
  g_free (chr_str);

  if (real_chr == (gchar) 182)
    gtk_text_thaw (text_wg);
}

/**********************************************************************
 * Like the previous, but also allows to go back and forth.
 */
void
forward_backward_cursor (gchar user_chr)
{
  gchar real_chr;
  gchar *chr_str;
  GtkText *text_wg;
  GtkEditable *edit_wg;

  text_wg = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (text_wg);

  if (user_chr == '\b')
    {
      g_timer_start (tmr_fast);
      if (cursor_pos < 3 || retro_pos == 0)
	{
	  tutor_beep ();
	  return;
	}

      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      gtk_text_forward_delete (text_wg, 1);
      gtk_text_insert (text_wg, font, &black, &blue, chr_str, 1);
      gtk_text_set_point (text_wg, cursor_pos);
      g_free (chr_str);

      cursor_pos--;
      retro_pos--;
      correcting++;

      if (cursor_pos == 2)
	{
	  gtk_text_set_point (text_wg, cursor_pos);
	  return;
	}

      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      if (chr_str[0] == ' ')
	{
	  g_free (chr_str);
	  chr_str =
	    gtk_editable_get_chars (edit_wg, cursor_pos - 1, cursor_pos);
	  if (chr_str[0] == '\n')
	    {
	      cursor_pos -= 2;
	      g_free (chr_str);
	      chr_str =
		gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
	      if (chr_str[0] == '\n')
		cursor_pos--;
	    }
	}
      g_free (chr_str);

      gtk_text_set_point (text_wg, cursor_pos);
      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      gtk_text_forward_delete (text_wg, 1);
      gtk_text_insert (text_wg, font, &black, &blue, chr_str, 1);
      gtk_text_set_point (text_wg, cursor_pos);
      g_free (chr_str);
    }
  else
    {
      if (cursor_pos >= (gtk_text_get_length (text_wg) - 2))
	{
	  g_timer_start (tmr_fast);
	  if (user_chr != '\n' || retro_pos)
	    return;
	}

      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      real_chr = chr_str[0];

      gtk_text_forward_delete (text_wg, 1);

      if ((user_chr == real_chr ||
	   ((real_chr == (gchar) 182) && user_chr == '\n')) && !retro_pos)
	{
	  n_touchs++;
	  tic_toc_time ();
	  if (!correcting)
	    gtk_text_insert (text_wg, font, &black, &green, chr_str, 1);
	  else
	    gtk_text_insert (text_wg, font, &black, &darkgreen, chr_str, 1);
	}
      else
	{
	  g_timer_start (tmr_fast);
	  tutor_beep ();
	  gtk_text_insert (text_wg, font, &black, &red, chr_str, 1);
	  n_errors++;
	  retro_pos++;
	}
      g_free (chr_str);

      if (correcting > 0)
	correcting--;

      cursor_blink = FALSE;
      cursor_pos++;

      if (real_chr == (gchar) 182)
	{
	  cursor_pos += 3;
	  if (cursor_pos < gtk_text_get_length (text_wg) - 3)
	    gtk_text_set_point (text_wg, cursor_pos);
	  else
	    {
	      return;
	    }
	}

      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      if (chr_str[0] == '\n')
	{
	  gtk_text_freeze (text_wg);
	  cursor_pos += 2;
	  gtk_text_set_point (text_wg, cursor_pos);
	  gtk_text_thaw (text_wg);
	}
      g_free (chr_str);

      cursor_blink = TRUE;

      if (real_chr == (gchar) 182)
	gtk_text_freeze (text_wg);

      chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
      gtk_text_forward_delete (text_wg, 1);
      gtk_text_insert (text_wg, font, &white, &black, chr_str, 1);
      gtk_text_set_point (text_wg, gtk_text_get_point (text_wg) - 1);
      g_free (chr_str);

      if (real_chr == (gchar) 182)
	gtk_text_thaw (text_wg);
    }
}

/**********************************************************************
 * Measures the elapsed time between the actual moment
 * and that of the last call.
 */
void
tic_toc_time (void)
{
  gulong micros;

  if (touch_time[0] < 2000 && touch_time[0] >= 0)
    {
      touch_time[0] += 1;
      touch_time[(int) touch_time[0]] = g_timer_elapsed (tmr_fast, &micros);
      g_timer_start (tmr_fast);
    }
}

/**********************************************************************
 * Initialize the cursor so that it blinks
 */
gboolean
init_cursor (gpointer data)
{
  cursor_blink = TRUE;
  cursor_is_initializing = FALSE;
  cursor_on (NULL);
  return FALSE;
}

/**********************************************************************
 * Turns on the cursor and makes it blink if 'cursor_blink' is TRUE
 */
gboolean
cursor_on (gpointer data)
{
  gchar *chr_str;
  GtkText *text_wg;
  GtkEditable *edit_wg;

  if (!cursor_blink)
    return FALSE;

  if (!GTK_IS_WIDGET (window_tutor_))
    {
      cursor_blink = FALSE;
      return FALSE;
    }

  text_wg = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (lookup_widget (window_tutor_, "text_tutor"));

  chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
  gtk_text_forward_delete (text_wg, 1);
  gtk_text_insert (text_wg, font, &white, &black, chr_str, 1);
  gtk_text_set_point (text_wg, gtk_text_get_point (text_wg) - 1);
  g_free (chr_str);

  g_timeout_add (TIME_CURSOR_ON, &cursor_off, NULL);

  return FALSE;
}

/**********************************************************************
 * Turns off the cursor and makes it blink if 'cursor_blink' is TRUE
 */
gboolean
cursor_off (gpointer data)
{
  gchar *chr_str;
  GtkText *text_wg;
  GtkEditable *edit_wg;

  if (!cursor_blink)
    return FALSE;

  if (!GTK_IS_WIDGET (window_tutor_))
    {
      cursor_blink = FALSE;
      return FALSE;
    }

  text_wg = (GtkText *) lookup_widget (window_tutor_, "text_tutor");
  edit_wg = (GtkEditable *) lookup_widget (window_tutor_, "text_tutor");

  chr_str = gtk_editable_get_chars (edit_wg, cursor_pos, cursor_pos + 1);
  gtk_text_forward_delete (text_wg, 1);
  gtk_text_insert (text_wg, font, &black, &blue, chr_str, 1);
  gtk_text_set_point (text_wg, gtk_text_get_point (text_wg) - 1);
  g_free (chr_str);

  g_timeout_add (TIME_CURSOR_OFF, &cursor_on, NULL);

  return FALSE;
}

/**********************************************************************
 * Charts manipulation
 */
gdouble y_data[30];

void
initialize_chart (GtkWidget * widget)
{
  gint i;
  GtkWidget *canvas;
  GtkWidget *plot;
  GtkWidget *dataset;
  static gdouble x_data[30];

  /*
   * Initialize X data
   */
  for (i = 0; i < 30; i++)
    x_data[i] = i + 1;

  /*
   * Canvas
   */
  canvas = gtk_plot_canvas_new (600, 280, 1.);
  GTK_PLOT_CANVAS_UNSET_FLAGS (GTK_PLOT_CANVAS (canvas),
			       GTK_PLOT_CANVAS_DND_FLAGS);
  gtk_widget_set_name (canvas, "canvas_stat");
  gtk_widget_ref (canvas);
  gtk_object_set_data_full (GTK_OBJECT (window_stat_), "canvas_stat", canvas,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_container_add (GTK_CONTAINER (widget), canvas);

  gtk_plot_canvas_set_background (GTK_PLOT_CANVAS (canvas), &blue);
  gtk_widget_show (canvas);

  /*
   * Plot
   */
  plot = gtk_plot_new_with_size (NULL, .90, .80);
  gtk_widget_set_name (plot, "plot_stat");
  gtk_widget_ref (plot);
  gtk_object_set_data_full (GTK_OBJECT (window_stat_), "plot_stat", plot,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_plot_canvas_add_plot (GTK_PLOT_CANVAS (canvas), GTK_PLOT (plot), .07,
			    .1);

  gtk_plot_clip_data (GTK_PLOT (plot), TRUE);
  gtk_plot_set_range (GTK_PLOT (plot), .5, 30.5, 0, 100);
  gtk_plot_axis_set_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_X, 1, 1);
  gtk_plot_axis_set_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_Y, 10, 0);
  gtk_plot_grids_set_visible (GTK_PLOT (plot), FALSE, TRUE, TRUE, FALSE);

  gtk_plot_axis_set_labels_numbers (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT, 0,
				    0);
  gtk_plot_axis_set_labels_numbers (GTK_PLOT (plot), GTK_PLOT_AXIS_BOTTOM, 0,
				    0);
  gtk_plot_axis_show_title (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT);
  gtk_plot_axis_move_title (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT, 0, .01, .06);
  gtk_plot_axis_set_title (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT, "Y");
  gtk_plot_axis_justify_title (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT,
			       GTK_JUSTIFY_LEFT);

  gtk_plot_axis_hide_title (GTK_PLOT (plot), GTK_PLOT_AXIS_BOTTOM);
  gtk_plot_axis_hide_title (GTK_PLOT (plot), GTK_PLOT_AXIS_TOP);
  gtk_plot_axis_hide_title (GTK_PLOT (plot), GTK_PLOT_AXIS_RIGHT);
  gtk_plot_axis_show_labels (GTK_PLOT (plot), GTK_PLOT_AXIS_RIGHT, 0);
  gtk_plot_axis_show_labels (GTK_PLOT (plot), GTK_PLOT_AXIS_TOP, 0);
  gtk_plot_axis_show_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_LEFT, 0, 0);
  gtk_plot_axis_show_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_BOTTOM, 0, 0);
  gtk_plot_axis_show_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_TOP, 0, 0);
  gtk_plot_axis_show_ticks (GTK_PLOT (plot), GTK_PLOT_AXIS_RIGHT, 0, 0);
  gtk_plot_hide_legends (GTK_PLOT (plot));

  gtk_plot_set_background (GTK_PLOT (plot), &white);
  gtk_widget_show (plot);

  /*
   * Data set
   */
  dataset = gtk_plot_data_new ();
  gtk_widget_set_name (dataset, "data_stat");
  gtk_widget_ref (dataset);
  gtk_object_set_data_full (GTK_OBJECT (window_stat_), "data_stat", dataset,
			    (GtkDestroyNotify) gtk_widget_unref);
  gtk_plot_add_data (GTK_PLOT (plot), GTK_PLOT_DATA (dataset));

  gtk_plot_data_set_symbol (GTK_PLOT_DATA (dataset), GTK_PLOT_SYMBOL_CIRCLE,
			    GTK_PLOT_SYMBOL_FILLED, 6, 2, &darkgreen, &black);
  gtk_plot_data_set_line_attributes (GTK_PLOT_DATA (dataset),
				     GTK_PLOT_LINE_SOLID, 2, &black);
  gtk_plot_data_set_x (GTK_PLOT_DATA (dataset), x_data);
  gtk_widget_show (dataset);

  plot_stat_log (1);
}

/**********************************************************************
 * Plots the statistics
 */
void
plot_stat_log (gint field)
{
  gint i;
  gchar *tmp_locale;
  gchar *tmp_name;
  gchar tmp_str[200];
  FILE *fh;
  gfloat accur[31];
  gfloat velo[31];
  gfloat fluid[31];
  gchar data[31][20];
  gchar hour[31][20];
  gchar lesson[31][20];
  GtkWidget *wg;

  /*
   * Select the file name
   */
  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/", NULL);
  switch (exercise_type)
    {
    case 1:
      tmp_name = g_strconcat (tmp_name, "stat_basic.txt", NULL);
      break;
    case 2:
      tmp_name = g_strconcat (tmp_name, "stat_adapt.txt", NULL);
      break;
    case 3:
      tmp_name = g_strconcat (tmp_name, "stat_velo.txt", NULL);
      break;
    case 4:
      tmp_name = g_strconcat (tmp_name, "stat_fluid.txt", NULL);
      break;
    }

  /*
   * Open the data file
   */
  fh = fopen (tmp_name, "r");
  if (!fh)
    {
      g_message ("coudn't find the statistic file:\n%s", tmp_name);
      g_free (tmp_name);
      return;
    }

  /*
   * Reads the first line (header)
   */
  if (!fgets (tmp_str, 200, fh))
    {
      g_message ("no data on the statistic file:\n%s", tmp_name);
      g_free (tmp_name);
      fclose (fh);
      return;
    }
  g_free (tmp_name);

  /*
   * Changing to "C" locale
   */
  tmp_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
  setlocale (LC_NUMERIC, "C");

  /*
   * Reads the first 30 points
   */
  i = 0;
  while (fscanf
	 (fh, "%f%f%f%s%s%s", &accur[i], &velo[i], &fluid[i], data[i],
	  hour[i], lesson[i]) == 6)
    {
      switch (field)
	{
	case 1:
	  y_data[i] = accur[i];
	  break;
	case 2:
	  y_data[i] = velo[i];
	  break;
	case 3:
	  y_data[i] = fluid[i];
	  break;
	}
      if (++i == 30)
	break;
    }

  /*
   * Reads until the end, keeping the last 30 points 
   */
  while (fscanf
	 (fh, "%f%f%f%s%s%s", &accur[i], &velo[i], &fluid[i], data[i],
	  hour[i], lesson[i]) == 6)
    {
      for (i = 0; i < 29; i++)
	y_data[i] = y_data[i + 1];
      switch (field)
	{
	case 1:
	  y_data[i] = accur[i + 1];
	  break;
	case 2:
	  y_data[i] = velo[i + 1];
	  break;
	case 3:
	  y_data[i] = fluid[i + 1];
	  break;
	}
      i++;
    }
  fclose (fh);

  /*
   * Coming back to the right locale
   */
  setlocale (LC_NUMERIC, tmp_locale);
  g_free (tmp_locale);

  /*
   * Plot the data
   */
  wg = lookup_widget (window_stat_, "data_stat");
  gtk_plot_data_set_y (GTK_PLOT_DATA (wg), y_data);
  gtk_plot_data_set_numpoints (GTK_PLOT_DATA (wg), i);

  /*
   * Formats the charts
   */
  wg = lookup_widget (window_stat_, "plot_stat");
  switch (field)
    {
    case 1:
      gtk_plot_set_range (GTK_PLOT (wg), .5, 30.5, 85, 100);
      gtk_plot_axis_set_ticks (GTK_PLOT (wg), GTK_PLOT_AXIS_Y, 1, 0);
      gtk_plot_axis_set_title (GTK_PLOT (wg), GTK_PLOT_AXIS_LEFT,
			       _("Accuracy (%)"));
      break;

    case 2:
      gtk_plot_set_range (GTK_PLOT (wg), .5, 30.5, 0, 120);
      gtk_plot_axis_set_ticks (GTK_PLOT (wg), GTK_PLOT_AXIS_Y, 10, 0);
      gtk_plot_axis_set_title (GTK_PLOT (wg), GTK_PLOT_AXIS_LEFT,
			       _("Velocity (WPM)"));
      break;

    case 3:
      gtk_plot_set_range (GTK_PLOT (wg), .5, 30.5, 0, 100);
      gtk_plot_axis_set_ticks (GTK_PLOT (wg), GTK_PLOT_AXIS_Y, 10, 0);
      gtk_plot_axis_set_title (GTK_PLOT (wg), GTK_PLOT_AXIS_LEFT,
			       _("Fluidness (%)"));
      break;
    }

  /*
   * Paint the result
   */
  wg = lookup_widget (window_stat_, "canvas_stat");
  gtk_plot_canvas_paint (GTK_PLOT_CANVAS (wg));
  gtk_plot_canvas_refresh (GTK_PLOT_CANVAS (wg));
}

/**********************************************************************
 * Compare two strings, so that it applies to other sorting functions.
 */
gint
comp_str (gconstpointer a, gconstpointer b)
{
  return (strcasecmp (a, b));
}
