/*
 * Set of functions to deal with internationalization (translation).
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "support.h"
#include "interface.h"

#include "global.h"
#include "velocity.h"
#include "fluidness.h"
#include "translation.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_main_;
extern gboolean callbacks_shield;
extern int _nl_msg_cat_cntr;
extern GList *word_list;
extern gint word_list_len;
extern Paragraph par;
extern gchar *kb_name;
extern gchar *data_path;

/**********************************************************************
 * Global variables used only here
 */
#include "languages.h"
const gchar *language_set = LANG_SET;
gchar lang[N_LANGS][50];
gchar code[N_LANGS][6];

/**********************************************************************
 * Initialize 'lang' and 'code' accordingly to 'language_set',
 * which was defined in "languages.h".
 */
void
set_lang_code ()
{
  gint i;
  gchar delimiter[2];
  gchar tmp_str[50];
  gchar **temp_lang = NULL;

  strcpy (delimiter, "\n");
  temp_lang = g_strsplit (language_set, delimiter, N_LANGS);
  strcpy (delimiter, "(");
  for (i = 0; i < N_LANGS; i++)
    {
      /*
       * Initialize 'lang'
       */
      strncpy (tmp_str, temp_lang[i], 50);
      tmp_str[50] = '\0';
      g_strdelimit (tmp_str, delimiter, '\0');
      strcpy (lang[i], g_strstrip (tmp_str));

      /*
       * Initialize 'code'
       */
      strcpy (tmp_str, strchr (temp_lang[i], '('));
      *(strchr (tmp_str, ')')) = '\0';
      strcpy (code[i], tmp_str + 1);
    }
  g_strfreev (temp_lang);
}

gboolean
estas_Esperanto ()
{
  return ((((unsigned) time (NULL)) % 10) == 1 ? TRUE : FALSE);
}

/**********************************************************************
 * Get the current locale and change it if necessary
 */
void
trans_init_language_env ()
{
  gchar tmp_code[33];
  gchar aux_code_2[3];
  gint i;

  /* 
   * Read the current locale and cut eventual colons
   */
  get_lang_env (tmp_code);
  if (tmp_code[2] == '_')
    tmp_code[5] = '\0';
  else
    tmp_code[2] = '\0';

  /*
   * Test if the current locale is available
   */
  set_lang_code ();
  for (i = 0; i < N_LANGS; i++)
    if (strcmp (code[i], tmp_code) == 0)
      break;

  /*
   * If not available, test for a similar one
   */
  if (i == N_LANGS)
    {
      for (i = 0; i < N_LANGS; i++)
	if (strstr (tmp_code, code[i]) != NULL)
	  {
	    strcpy (tmp_code, code[i]);
	    break;
	  }
    }

  /*
   * If even a similar is not available, then change it
   */
  if (i == N_LANGS)
    {
      g_message ("as the locale \"%s\" isn't available,", tmp_code);
      strcpy (aux_code_2, tmp_code);
      aux_code_2[2] = '\0';
      if (strcmp (aux_code_2, "pt") == 0)
	strcpy (tmp_code, "pt_BR");

      /* this second is just a template */
      else if (strcmp (tmp_code, "pt_PT") == 0)
	strcpy (tmp_code, "pt_BR");

      else if (strcmp (aux_code_2, "es") == 0)
	strcpy (tmp_code, "pt_BR");

      else if (estas_Esperanto () == TRUE)
	strcpy (tmp_code, "eo_EO");

      else
	strcpy (tmp_code, "C");

      g_message ("we are using this another one: \"%s\"", tmp_code);
      g_message ("if you don't agree, try to change it at the main menu...");
    }

  setenv ("LANGUAGE", tmp_code, 1);
  /*++_nl_msg_cat_cntr; */
  gtk_set_locale ();
}

/**********************************************************************
 * Inserts the list of available languages in the 'combo_language'.
 */
void
set_combo_language ()
{
  gint i;
  gint i_env;
  gchar tmp_code[12];
  GList *languages = NULL;
  GtkWidget *wg;

  get_lang_env (tmp_code);

  for (i = 0, i_env = -1; i < N_LANGS; i++)
    {
      if (strcmp (code[i], tmp_code) != 0)
	languages = g_list_insert_sorted (languages, lang[i], &comp_str);
      else
	i_env = i;
    }

  if (i_env == -1)
    g_error ("set_combo_language() ==> the locale \"%s\" is not available!",
	     tmp_code);
  else
    languages = g_list_insert (languages, lang[i_env], 0);

  wg = lookup_widget (window_main_, "combo_language");
  gtk_combo_set_popdown_strings (GTK_COMBO (wg), languages);

  g_list_free (languages);
}

/**********************************************************************
 * Update the current language used accordingly to that selected in the
 * 'combo_language'
 */
void
change_language ()
{
  gint i;
  gint pos = 0;
  gchar *selection = NULL;
  GtkWidget *wg;

  wg = lookup_widget (window_main_, "entry_language");
  selection = gtk_editable_get_chars (GTK_EDITABLE (wg), 0, -1);

  for (i = 0; i < N_LANGS; i++)
    if (strcmp (lang[i], selection) == 0)
      break;

  if (i == N_LANGS)
    {
      g_warning ("change_language() --> couldn't find the language: %s",
		 selection);
      g_free (selection);
      return;
    }
  g_free (selection);

  setenv ("LANGUAGE", code[i], 1);
  /*++_nl_msg_cat_cntr; */
  gtk_set_locale ();

  callbacks_shield = TRUE;
  gtk_widget_destroy (window_main_);
  window_main_ = create_window_main ();
  wg = lookup_widget (window_main_, "entry_keyboard");
  gtk_editable_insert_text (GTK_EDITABLE (wg), kb_name, strlen (kb_name),
			    &pos);
  set_combo_language ();
  gtk_widget_show (window_main_);
  callbacks_shield = FALSE;

  g_list_free (word_list);
  word_list = NULL;
  word_list_len = 0;
  g_free (par.str);
  par.str = NULL;
  par.n = 0;

  g_message ("selected language code --> %s", code[i]);
}

/**********************************************************************
 * Reads the text of a file to be presented at some dialog,
 * accordingly to the environment language.
 * 'text': buffer where the text is copied into.
 * 'max_text_len': the size of the buffer.
 * 'file_end': how the file name ends (isn't needed the leading
 * underscore, only the termination, ex.: 'basic_end.txt').
 */
void
read_text (gchar * text, gint max_text_len, const gchar * file_end)
{
  gchar *tmp_name = NULL;
  gchar *tmp_path = NULL;
  gchar tmp_code[16];
  FILE *fh;

  get_lang_env (tmp_code);
  if (strlen (tmp_code) > 5)
    tmp_code[5] = '\0';
  tmp_name = g_strconcat (tmp_code, "_", file_end, NULL);

  /*
   * Try at home
   */
  tmp_path = g_strconcat (g_get_home_dir (), ".klavaro/", tmp_name, NULL);
  fh = fopen (tmp_path, "r");

  /*
   * Try at PACKAGE
   */
  if (fh == NULL)
    {
      g_free (tmp_path);
      tmp_path = g_strconcat (data_path, tmp_name, NULL);
      fh = fopen (tmp_path, "r");
    }

  /*
   * Default to C
   */
  if (fh == NULL && strcmp (tmp_code, "C") != 0)
    {
      g_message ("read_text() --> couldn't open the data file: %s\n"
		 " So, we have to apply the default one: C_%s", tmp_name,
		 file_end);
      g_free (tmp_name);
      g_free (tmp_path);
      setenv ("LANGUAGE", "C", 1);
      read_text (text, max_text_len, file_end);
      setenv ("LANGUAGE", tmp_code, 1);
      return;
    }

  if (fh == NULL)
    g_error ("read_text() --> couldn't open the data file:\n %s", tmp_name);
  else
    {
      gint i, j, tmpc = 0;

      max_text_len -= 2;
      i = 0;
      do
	{
	  for (j = 0; j < 60; j++)
	    {
	      tmpc = fgetc (fh);
	      text[i] = (tmpc != EOF) ? (gchar) tmpc : '\0';
	      if (i <= max_text_len)
		i++;
	      if (tmpc == '\n')
		break;
	    }
	  if (tmpc == '\n')
	    continue;
	  for (j = 0; j < 20; j++)
	    {
	      tmpc = fgetc (fh);
	      if (tmpc == ' ')
		{
		  text[i] = '\n';
		  break;
		}
	      else if (tmpc == '\n')
		{
		  text[i] = '\n';
		  break;
		}
	      else if (tmpc == EOF)
		{
		  text[i] = '\0';
		  break;
		}
	      text[i] = (gchar) tmpc;
	      if (i < max_text_len)
		i++;
	    }
	  i++;
	}
      while (i < max_text_len && tmpc != EOF);
      fclose (fh);
      text[max_text_len - 1] = '\0';
    }

  g_free (tmp_name);
  g_free (tmp_path);
}
