/*
 * Velocity exercise
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <libgen.h>
#include <unistd.h>

#include "support.h"

#include "global.h"
#include "callbacks.h"
#include "translation.h"
#include "keyboard.h"
#include "velocity.h"

/**********************************************************************
 * Global variables
 */
extern GtkWidget *window_tutor_;
extern gchar *data_path;
extern gboolean cursor_is_initializing;
extern gboolean cursor_blink;
extern gint tutor_query;
extern gint cursor_pos;
extern gint n_touchs;
extern gint n_errors;
extern gdouble elapsed_time;
extern GTimer *tmr;
extern GdkFont *font;
extern GdkColor black;
extern GdkColor white;
extern GList *word_list;
extern gint word_list_len;

/**********************************************************************
 * Global variables only used here
 */
gchar word_list_name[21];

/**********************************************************************
 * Retrieves words from the dictionary.
 */
void
init_word_list (gchar * list_name)
{
  static gchar *word_str = NULL;	/* no doubts: this is initialized only
					   at the first call */
  gchar *tmp_name;
  gchar tmp_code[16];
  gchar str_32[33];
  FILE *fh;

  tmp_name =
    g_strconcat (g_get_home_dir (), "/.klavaro/last_word_list", NULL);
  if (list_name && strcmp (list_name, _("--> Default")))
    {
      assert_user_dir ();
      if ((fh = fopen (tmp_name, "w")))
	{
	  fprintf (fh, "%s", list_name);
	  fclose (fh);
	}
      else
	g_message ("could not write the file:\n %s", tmp_name);

      g_free (tmp_name);
      tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/",
			      list_name, ".words", NULL);
      g_message ("loading dictionary: %s.words", list_name);
      strncpy (word_list_name, list_name, 20);
      word_list_name[20] = '\0';
    }
  else
    {
      unlink (tmp_name);
      g_free (tmp_name);
      get_lang_env (tmp_code);
      tmp_name = g_strconcat (data_path, tmp_code, ".words", NULL);
      g_message ("loading dictionary: %s.words", tmp_code);
      strcpy (word_list_name, "Default");
    }

  if ((fh = fopen (tmp_name, "r")))
    {
      g_free (word_str);
      g_list_free (word_list);
      word_list = NULL;
      word_str = g_strconcat ("", NULL);
      while (fgets (str_32, 33, fh))
	{
	  str_32[strlen (str_32) - 1] = '\0';
	  word_str = g_strconcat (word_str, " ", str_32, NULL);
	  word_list = g_list_append (word_list, strrchr (word_str, ' ') + 1);
	}
      fclose (fh);
      word_list_len = g_list_length (word_list);
      g_message ("dictionary loaded!\n");
    }
  else
    {
      g_message ("could not open the file: %s", tmp_name);
      g_free (tmp_name);
      get_lang_env (tmp_code);
      if (strcmp (tmp_code, "C") == 0)
	g_error ("so, we must quit!");
      setenv ("LANGUAGE", "C", 1);
      init_word_list (list_name);
      setenv ("LANGUAGE", tmp_code, 1);
      return;
    }

  g_free (tmp_name);
}

/**********************************************************************
 * Initialize the velo exercise window.
 */
void
velo_init ()
{
  gchar *tmp_name;
  gchar wl_name[1000];
  FILE *fh;

  gtk_window_set_title (GTK_WINDOW (window_tutor_), _("Klavaro - Velocity"));
  gtk_widget_hide (lookup_widget (window_tutor_, "frame_lesson"));
  gtk_widget_hide (lookup_widget (window_tutor_, "vseparator_tutor"));
  gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_show"));
  gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_other"));

  gtk_label_set_text (GTK_LABEL
		      (lookup_widget (window_tutor_, "label_heading")),
		      _("Velocity exercises: accelerate typing real words."));

  if (word_list_len == 0)
    {
      tmp_name =
	g_strconcat (g_get_home_dir (), "/.klavaro/last_word_list", NULL);
      if ((fh = fopen (tmp_name, "r")))
	{
	  fscanf (fh, "%s", wl_name);
	  g_free (tmp_name);
	  tmp_name =
	    g_strconcat (g_get_home_dir (), "/.klavaro/", wl_name, ".words",
			 NULL);
	  fclose (fh);
	  if ((fh = fopen (tmp_name, "r")))
	    {
	      fclose (fh);
	      init_word_list (wl_name);
	    }
	}
      g_free (tmp_name);
    }
  if (word_list_len == 0)
    init_word_list (NULL);

  tutor_query = QUERY_INTRO;
  velo_update ();
}

/**********************************************************************
 * Update what is shown in the tutor window. 
 */
void
velo_update ()
{
  gchar text[4096];
  GtkText *wg_text;
  GtkEditable *edit_wg;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  edit_wg = GTK_EDITABLE (lookup_widget (window_tutor_, "text_tutor"));

  gtk_text_freeze (wg_text);
  switch (tutor_query)
    {
    case QUERY_INTRO:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      read_text (text, 4096, "velo_intro.txt");
      gtk_text_insert (wg_text, font, &black, &white, text, -1);
      tutor_message (_("Press any key to start the exercise. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_START:
      gtk_text_set_point (wg_text, 0);
      gtk_text_forward_delete (wg_text, gtk_text_get_length (wg_text));
      draw_random_words ();
      tutor_message (_("Start typing when you are ready. "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_PROCESS_TOUCHS:
      tutor_message (_("STARTED! => "));
      gtk_widget_show (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;

    case QUERY_END:
      tutor_message (_("<<==  End of exercise."
		       " Press [Enter] to start another. ==>> "));
      gtk_widget_hide (lookup_widget (window_tutor_, "button_tutor_restart"));
      break;
    }
  gtk_text_thaw (wg_text);
}

/**********************************************************************
 * Responds to each touch of the user, according to the 'tutor_query' mode
 */
void
velo_process_touch (gchar user_chr)
{
  GtkText *wg_text;

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  switch (tutor_query)
    {
    case QUERY_PROCESS_TOUCHS:
      break;

    case QUERY_INTRO:
      tutor_query = QUERY_START;
      velo_update ();
      n_touchs = 0;
      n_errors = 0;
      gtk_text_set_point (wg_text, 2);
      cursor_pos = 2;
      if (!cursor_is_initializing)
	{
	  cursor_blink = FALSE;
	  cursor_is_initializing = TRUE;
	  g_timeout_add (TIME_CURSOR_ON + 100, &init_cursor, NULL);
	}
      return;

    case QUERY_START:
      tutor_query = QUERY_PROCESS_TOUCHS;
      velo_update ();
      g_timer_start (tmr);
      forward_cursor (user_chr);
      return;

    case QUERY_END:
      if (user_chr == '\n')
	{
	  tutor_query = QUERY_INTRO;
	  velo_process_touch ('\0');
	}
      else
	{
	  tutor_beep ();
	  velo_update ();
	}
      return;

    default:
      gdk_beep ();
      g_warning ("velo_process_touch(): the program mustn't pass here!");
      return;
    }

  /*
   * It is time to analise the correctness of typing.
   */
  forward_cursor (user_chr);
  if (cursor_pos >= gtk_text_get_length (wg_text))
    {
      elapsed_time = g_timer_elapsed (tmr, NULL);
      velo_end_exercise ();
      tutor_query = QUERY_END;
      velo_update ();
      tutor_beep ();
      return;
    }
}

/**********************************************************************
 * Put on the screen the final results
 */
#define MAX_TEXT_LEN 1024
void
velo_end_exercise ()
{
  gchar *tmp_locale;
  gchar *str;
  gchar tmp_str[MAX_TEXT_LEN];
  gchar *tmp_name;
  GtkText *wg_text;
  FILE *fh;
  time_t tmp_time;
  struct tm *ltime;
  gint minutes;
  gint seconds;
  gdouble accuracy;
  gdouble touchs_per_second;
  gdouble words_per_minute;

  /*
   * Calculate statistics
   */
  minutes = ((gulong) elapsed_time) / 60;
  seconds = ((gulong) elapsed_time) % 60;

  accuracy = 100 * (1.0 - (gfloat) n_errors / n_touchs);
  touchs_per_second = (n_touchs - n_errors) / elapsed_time;
  words_per_minute = 10 * touchs_per_second;

  /*
   * Changing to "C" locale
   */
  tmp_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
  setlocale (LC_NUMERIC, "C");

  /*
   * Logging
   */
  tmp_name = g_strconcat (g_get_home_dir (), "/.klavaro/stat_velo.txt", NULL);
  assert_user_dir ();
  if (!(fh = fopen (tmp_name, "r")))
    {
      fh = fopen (tmp_name, "w");
      fprintf (fh, "Accuracy\tVelocity\tFluidness\tDate\tHour\tLesson\n");
    }
  else
    {
      fclose (fh);
      fh = fopen (tmp_name, "a");
    }

  if (fh)
    {
      tmp_time = time (NULL);
      ltime = localtime (&tmp_time);
      fprintf (fh, "%.2f\t%.2f\t%.2f\t%i-%2.2i-%2.2i\t%2.2i:%2.2i\t%s\n",
	       accuracy, words_per_minute, 0.0, (ltime->tm_year) + 1900,
	       (ltime->tm_mon) + 1, (ltime->tm_mday), (ltime->tm_hour),
	       (ltime->tm_min), word_list_name);
      fclose (fh);
    }
  else
    g_message ("not able to log on this file:\n %s", tmp_name);
  g_free (tmp_name);

  /*
   * Coming back to the right locale
   */
  setlocale (LC_NUMERIC, tmp_locale);
  g_free (tmp_locale);

  /*
   * Print statistics
   */
  str = g_strconcat ("\n\n ", _("STATISTICS"), "\n\n ",
		     _("Elapsed time:"), " %i ",
		     _("minute(s) and"), " %i ", _("seconds"), "\n\n ",
		     _("Accuracy:"), " %.1f %%\t\t",
		     _("Goal:"), " %i %%", "\n\n ",
		     _("Characters per minute:"), " %.2f\t",
		     _("Goal:"), " %i ", _("(CPS)"), "\n ",
		     _("Words per minute:"), " %.1f\t",
		     _("Goal:"), " %i ", _("(WPM)"), "\n\n", NULL);
  sprintf (tmp_str, str, minutes, seconds, accuracy, 97,
	   words_per_minute / 10, 5, words_per_minute, 50);
  g_free (str);

  wg_text = GTK_TEXT (lookup_widget (window_tutor_, "text_tutor"));
  gtk_text_insert (wg_text, font, &black, &white, tmp_str, -1);

  /*
   * Comments
   */
  strcpy (tmp_str, _(" Comments:\n\n"));
  str = tmp_str + strlen (tmp_str);
  strcpy (str, "...");
  if (accuracy < 97)
    sprintf (str,
	     _(" Your accuracy rate is too low.\n"
	       " I only can rely on you if it is greater than 97 %%.\n"
	       " Please do it again.\n"));
  else if (words_per_minute < 10)
    sprintf (str,
	     _(" You must be kidding. You are so slow!\n"
	       " Please do it again and again.\n"));
  else if (words_per_minute < 20)
    sprintf (str,
	     _(" Still away from the highway. You can make better...\n"
	       " Try at least 20 WPM.\n"));
  else if (words_per_minute < 30)
    sprintf (str,
	     _(" You are doing well, but need to go faster.\n"
	       " And don't forget the accuracy. Try to get 30 WPM.\n"));
  else if (words_per_minute < 40)
    sprintf (str,
	     _(" Fine. Now you need to start running.\n"
	       " Can you reach 40 WPM?\n"));
  else if (words_per_minute < 50)
    sprintf (str, _(" Very good. You are almost there.\n"
		    " Can you finally reach 50 WPM?\n"));
  else if (words_per_minute < 60)
    sprintf (str, _(" Excellent. For this course, that is enough.\n"
		    " Try now the fluidness exercises, OK?\n"));
  else if (words_per_minute < 70)
    sprintf (str, _(" Fast! Are you training for a competition?\n"
		    " So, try to get 70 WPM!\n"));
  else if (words_per_minute < 80)
    sprintf (str,
	     _(" Top of \"qwerty\"."
	       " Now it's time to change to the Dvorak mode.\n"
	       " Are you afraid of reaching 80 WPM?\n"));
  else if (words_per_minute < 90)
    sprintf (str, _(" Dvorak mode dominated!\n" " Can you fly at 90 WPM?\n"));
  else
    sprintf (str,
	     _(" Dvorak master!\n"
	       " I have no words to express my admiration!\n"));

  gtk_text_insert (wg_text, NULL, &black, &white, tmp_str, -1);
}

/**********************************************************************
 * Draw random words selected from a 'discretionary'
 */
void
draw_random_words ()
{
  gchar *text;
  gchar str_32[33];
  gint i, j;

  for (i = 0; i < 4; i++)
    {				/* 4 paragraphs per exercise */
      text = NULL;
      for (j = 0; j < 20; j++)
	{			/* 20 words per paragraph */
	  get_random_word (str_32);
	  if (!j)
	    {
	      str_32[0] = toupper (str_32[0]);
	      text = g_strconcat (str_32, NULL);
	    }
	  else
	    text = g_strconcat (text, " ", str_32, NULL);
	}
      text = g_strconcat (text, ".\n", NULL);
      draw_paragraph (text);
      g_free (text);
    }
}

/**********************************************************************
 * Selects one word from a 'discretionary'
 */
void
get_random_word (gchar str_32[33])
{
  strncpy (str_32, g_list_nth_data (word_list, rand () % word_list_len), 33);
  str_32[32] = '\0';
}

/**********************************************************************
 * Reads the text file "file_name" and write to the dictionary.
 * If overwrite is TRUE, then besides overwriting any .words file,
 * it loads a lesson with the new dictionary.
 */
#define MAX_WORD_LEN 30
void
velo_create_dict (gchar * file_name, gboolean overwrite)
{
  gint i;
  gchar tmp_chr;
  gchar tmp_word[MAX_WORD_LEN];
  gchar *dict_path;
  gchar *dict_name;
  FILE *fh_source;
  FILE *fh_destiny;

  if (!file_name)
    {
      gdk_beep ();
      g_warning ("velo_create_dict(): null file name as first argument.");
      return;
    }

  if (!(fh_source = fopen (file_name, "r")))
    {
      gdk_beep ();
      g_warning ("couldn't open the file:\n %s\n", file_name);
      return;
    }

  dict_path =
    g_strconcat (g_get_home_dir (), "/.klavaro/", basename (file_name), NULL);
  if (strrchr (strrchr (dict_path, '/'), '.'))
    *(strrchr (dict_path, '.')) = '\0';
  dict_name = g_strconcat (basename (dict_path), NULL);
  dict_path = g_strconcat (dict_path, ".words", NULL);

  assert_user_dir ();
  if (!(fh_destiny = fopen (dict_path, "w")))
    {
      gdk_beep ();
      g_warning ("couldn't create the file:\n %s\n", dict_path);
      if (!overwrite)
	{
	  fclose (fh_source);
	  g_free (dict_path);
	  g_free (dict_name);
	  return;
	}
    }
  g_free (dict_path);

  i = 0;
  while (1)
    {
      tmp_chr = getc (fh_source);
      if (!(isalpha (tmp_chr)))
	{
	  if (!i && tmp_chr != EOF)
	    continue;
	  tmp_word[i] = '\0';
	  if (!i)
	    break;

	  if (strlen (tmp_word) > 1)
	    {
	      fputs (tmp_word, fh_destiny);
	      if (tmp_chr == EOF)
		break;
	      fputc ('\n', fh_destiny);
	    }
	  if (tmp_chr == EOF)
	    break;
	  i = 0;
	  continue;
	}
      if (i < (MAX_WORD_LEN - 1))
	{
	  tmp_word[i] = tmp_chr;
	  i++;
	}
    }
  fclose (fh_source);
  fclose (fh_destiny);

  if (overwrite)
    {
      init_word_list (dict_name);
      tutor_query = QUERY_INTRO;
      velo_process_touch ('\0');
    }

  g_free (dict_name);
}
